"""Contains several functions for crop farmer module."""

import numpy as np
import numpy.typing as npt
from numba import njit, prange

from geb.geb_types import (
    ArrayBool,
    ArrayFloat32,
    ArrayInt32,
    ThreeDArrayInt32,
    TwoDArrayFloat32,
    TwoDArrayInt32,
)
from geb.hydrology.landsurface.evapotranspiration import get_root_ratios


@njit(cache=True, inline="always")
def get_farmer_HRUs(
    field_indices: npt.NDArray[np.int32],
    field_indices_by_farmer: npt.NDArray[np.int32],
    farmer_index: int,
) -> npt.NDArray[np.int32]:
    """Get HRU indices for a given farmer.

    Args:
        field_indices: Flat array of HRU indices for all fields (ordered by farmer).
        field_indices_by_farmer: Start/stop indices into ``field_indices`` per farmer.
        farmer_index: Farmer index to select.

    Returns:
        HRU indices that belong to ``farmer_index``.
    """
    return field_indices[
        field_indices_by_farmer[farmer_index, 0] : field_indices_by_farmer[
            farmer_index, 1
        ]
    ]


@njit(cache=True)
def farmer_command_area(
    n: int,
    field_indices: npt.NDArray[np.int32],
    field_indices_by_farmer: npt.NDArray[np.int32],
    reservoir_command_areas: ArrayInt32,
) -> ArrayInt32:
    """Map each farmer to the first available reservoir command area.

    Args:
        n: Number of farmers.
        field_indices: Flat array of HRU indices for all fields (ordered by farmer).
        field_indices_by_farmer: Start/stop indices into ``field_indices`` per farmer.
        reservoir_command_areas: Command-area id per HRU (``-1`` if none).

    Returns:
        Array of length ``n`` with command-area id per farmer (``-1`` if none).
    """
    output = np.full(n, -1, dtype=np.int32)
    for farmer_i in range(n):
        farmer_fields = get_farmer_HRUs(
            field_indices, field_indices_by_farmer, farmer_i
        )
        for field in farmer_fields:
            command_area = reservoir_command_areas[field]
            if command_area != -1:
                output[farmer_i] = command_area
                break
    return output


@njit(cache=True)
def get_farmer_groundwater_depth(
    n: int,
    groundwater_depth: npt.NDArray[np.float32],
    HRU_to_grid: npt.NDArray[np.int32],
    field_indices: npt.NDArray[np.int32],
    field_indices_by_farmer: npt.NDArray[np.int32],
    cell_area: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Compute area-weighted mean groundwater depth per farmer.

    Args:
        n: Number of farmers.
        groundwater_depth: Groundwater depth per grid cell.
        HRU_to_grid: Map from HRU index to grid-cell index.
        field_indices: Flat array of HRU indices for all fields (ordered by farmer).
        field_indices_by_farmer: Start/stop indices into ``field_indices`` per farmer.
        cell_area: HRU area.

    Returns:
        Area-weighted groundwater depth per farmer.
    """
    groundwater_depth_by_farmer = np.full(n, np.nan, dtype=np.float32)
    for farmer_i in range(n):
        farmer_fields = get_farmer_HRUs(
            field_indices=field_indices,
            field_indices_by_farmer=field_indices_by_farmer,
            farmer_index=farmer_i,
        )
        total_cell_area = 0.0
        total_groundwater_depth_times_area = 0.0
        for field in farmer_fields:
            grid_cell = HRU_to_grid[field]
            total_cell_area += cell_area[field]
            total_groundwater_depth_times_area += (
                groundwater_depth[grid_cell] * cell_area[field]
            )
        groundwater_depth_by_farmer[farmer_i] = (
            total_groundwater_depth_times_area / total_cell_area
        )
    return groundwater_depth_by_farmer


@njit(cache=True, inline="always")
def get_deficit_between_dates(
    cumulative_water_deficit_m3: TwoDArrayFloat32,
    farmer: int,
    start_index: int,
    end_index: int,
) -> np.float32:
    """Get the water deficit between two dates for a farmer.

    Args:
        cumulative_water_deficit_m3: Cumulative water deficit in m3 for each day of the year for each farmer.
        farmer: The index of the farmer in the cumulative water deficit.
        start_index: Start day of cumulative water deficit calculation (index-based; Jan 1 == 0).
        end_index: End day of cumulative water deficit calculation (index-based; Jan 1 == 0).

    Returns:
        Water deficit in m3 between the two dates.

    Raises:
        ValueError: If deficit is negative.
    """
    if end_index == start_index:
        deficit = np.float32(0)
    elif end_index > start_index:
        deficit: np.float32 = (
            cumulative_water_deficit_m3[farmer, end_index]
            - cumulative_water_deficit_m3[
                farmer, start_index
            ]  # current day of year is effectively starting "tommorrow" due to Python's 0-indexing
        )
    else:  # end < start
        deficit: np.float32 = cumulative_water_deficit_m3[farmer, -1] - (
            cumulative_water_deficit_m3[farmer, start_index]
            - cumulative_water_deficit_m3[farmer, end_index]
        )

    if deficit < np.float32(0):
        raise ValueError("Deficit must be positive or zero")
    return deficit


@njit(cache=True)
def get_future_deficit(
    farmer: int,
    day_index: int,
    cumulative_water_deficit_m3: npt.NDArray[np.float32],
    crop_calendar: npt.NDArray[np.int32],
    crop_rotation_year_index: npt.NDArray[np.int32],
    potential_irrigation_consumption_farmer_m3: np.float32,
    reset_day_index: int,
) -> np.float32:
    """Get the future water deficit for a farmer.

    Args:
        farmer: Farmer index.
        day_index: Current day index (0-indexed).
        cumulative_water_deficit_m3: Cumulative water deficit in m3 for each day of the year for each farmer.
        crop_calendar: Crop calendar for each farmer. Each row is a farmer, and each column is a crop.
            Each crop is a list of [crop_type, planting_day, growing_days, crop_year_index].
            Planting day is 0-indexed (Jan 1 == 0).
            Growing days is the number of days the crop grows.
            Crop year index is the index of the year in the crop rotation.
        crop_rotation_year_index: Crop rotation year index for each farmer.
        potential_irrigation_consumption_farmer_m3: Potential irrigation consumption in m3 for each farmer on the current day.
        reset_day_index: Day index to reset the water year (0-indexed; Jan 1 == 0). Default is 0. Deficit
            is calculated up to this day. For example, when the reset day index is 364, the
            deficit is calculated up to Dec 31. When the reset day index is 0, the deficit is
            calculated up to Jan 1.

    Returns:
        Future water deficit in m3 for the farmer in the growing season.

    Raises:
        ValueError: If reset_day_index is not between 0 and 364 inclusive.
        ValueError: If start_day_index is not between 0 and 364 inclusive.
        ValueError: If future_water_deficit is negative.
    """
    if reset_day_index >= 365 or reset_day_index < 0:
        raise ValueError("Reset day index must be lower than 365 and greater than -1")
    future_water_deficit: np.float32 = potential_irrigation_consumption_farmer_m3
    for crop in crop_calendar[farmer]:
        crop_type = crop[0]
        crop_year_index = crop[3]
        if crop_type != -1 and crop_year_index == crop_rotation_year_index[farmer]:
            start_day_index = crop[1]
            if start_day_index < 0 or start_day_index >= 365:
                raise ValueError("Start day must be lower than 365 and greater than -1")
            growth_length = crop[2]

            relative_start_day_index = (start_day_index - reset_day_index) % 365
            relative_end_day_index = relative_start_day_index + growth_length
            relative_day_index = (day_index - reset_day_index) % 365

            if relative_end_day_index > 365:
                relative_end_day_index = 365

            if relative_start_day_index < relative_day_index:
                relative_start_day_index = relative_day_index

            if relative_start_day_index == relative_day_index:
                relative_start_day_index = (relative_start_day_index + 1) % 365

            if relative_day_index >= relative_end_day_index:
                continue
            else:
                future_water_deficit += get_deficit_between_dates(
                    cumulative_water_deficit_m3,
                    farmer,
                    (relative_start_day_index + reset_day_index) % 365,
                    (relative_end_day_index + reset_day_index) % 365,
                )

    if future_water_deficit < 0:
        raise ValueError("Future water deficit must be positive or zero")
    return future_water_deficit


@njit(cache=True)
def adjust_irrigation_to_limit(
    farmer: int,
    day_index: int,
    remaining_irrigation_limit_m3: np.ndarray,
    cumulative_water_deficit_m3: np.ndarray,
    crop_calendar: np.ndarray,
    crop_rotation_year_index: np.ndarray,
    farmer_gross_irrigation_demand_m3: float,
    irrigation_efficiency_farmer: float,
    reset_day_index: np.int32,
) -> np.float32:
    """Adjusts the irrigation to the remaining irrigation limit for a farmer.

    Args:
        farmer: Index of the farmer.
        day_index: Current day index (0-indexed).
        remaining_irrigation_limit_m3: Remaining irrigation limit in m3 for each farmer.
        cumulative_water_deficit_m3: Cumulative water deficit in m3 for each day of the year for each farmer.
        crop_calendar: Crop calendar for each farmer. Each row is a farmer, and each column is a crop.
        crop_rotation_year_index: Crop rotation year index for each farmer. Used to select the correct crop calendar.
        farmer_gross_irrigation_demand_m3: Gross irrigation demand in m3 for the farmer on the current day.
        irrigation_efficiency_farmer: Irrigation efficiency for the farmer.
        reset_day_index: Day to reset the irrigation year, effectively the end of the decision period.

    Returns:
        irrigation_correction_factor: The ratio of the remaining irrigation limit to the future water deficit.
        If the remaining irrigation limit is less than 0, returns 0.
    """
    if remaining_irrigation_limit_m3[farmer] < np.float32(0):
        return np.float32(0)
    # calculate future water deficit, but also include today's irrigation consumption
    # Check whether a full year has passed
    if np.all(~np.isnan(cumulative_water_deficit_m3[farmer])):
        potential_irrigation_consumption_farmer_m3 = (
            farmer_gross_irrigation_demand_m3 * irrigation_efficiency_farmer
        )
        future_water_deficit = get_future_deficit(
            farmer=farmer,
            day_index=day_index,
            cumulative_water_deficit_m3=cumulative_water_deficit_m3,
            crop_calendar=crop_calendar,
            crop_rotation_year_index=crop_rotation_year_index,
            potential_irrigation_consumption_farmer_m3=potential_irrigation_consumption_farmer_m3,
            reset_day_index=reset_day_index,
        )
        effective_remaining_irrigation_limit_m3 = (
            remaining_irrigation_limit_m3[farmer] * irrigation_efficiency_farmer
        )

        irrigation_correction_factor: np.float32 = (
            effective_remaining_irrigation_limit_m3 / future_water_deficit
        )
        # limit the ratio to 1 if the deficit is smaller than the limit
        irrigation_correction_factor: np.float32 = min(
            irrigation_correction_factor, np.float32(1)
        )
    else:
        irrigation_correction_factor: np.float32 = np.float32(1)

    assert future_water_deficit > np.float32(0)
    return irrigation_correction_factor


@njit(cache=True)
def withdraw_channel(
    available_channel_storage_m3: npt.NDArray[np.float32],
    grid_cell: int,
    cell_area: npt.NDArray[np.float32],
    field: int,
    farmer: int,
    irrigation_water_demand_field_m: float,
    water_withdrawal_m: npt.NDArray[np.float32],
    remaining_irrigation_limit_m3: npt.NDArray[np.float32],
    channel_abstraction_m3_by_farmer: npt.NDArray[np.float32],
    maximum_abstraction_channel_m3_field: np.float32,
    minimum_channel_storage_m3: float = 100.0,
) -> float:
    """Withdraw irrigation water from channel for a field.

    Args:
        available_channel_storage_m3: Available channel storage by grid cell (m³).
        grid_cell: Grid-cell index of the field.
        cell_area: HRU area (m²).
        field: HRU index.
        farmer: Farmer index.
        irrigation_water_demand_field_m: Remaining demand at field (m).
        water_withdrawal_m: Per-field withdrawal accumulator (m).
        remaining_irrigation_limit_m3: Remaining seasonal limit per farmer (m³),
            or ``NaN`` when unlimited.
        channel_abstraction_m3_by_farmer: Per-farmer channel abstraction accumulator (m³).
        maximum_abstraction_channel_m3_field: Max channel abstraction allowed for this
            field in current step (m³).
        minimum_channel_storage_m3: Minimum storage to keep in channel (buffer).

    Returns:
        Updated remaining field demand (m).
    """
    water_demand_cell_m3 = irrigation_water_demand_field_m * cell_area[field]
    assert water_demand_cell_m3 >= 0

    remaining_channel_m3 = max(
        available_channel_storage_m3[grid_cell] - minimum_channel_storage_m3, 0
    )

    # channel abstraction
    channel_abstraction_cell_m3 = min(
        remaining_channel_m3,
        water_demand_cell_m3,
        maximum_abstraction_channel_m3_field,
    )
    # ensure channel abstraction is non-negative
    assert channel_abstraction_cell_m3 >= 0

    available_channel_storage_m3[grid_cell] -= channel_abstraction_cell_m3
    channel_abstraction_cell_m = channel_abstraction_cell_m3 / cell_area[field]
    water_withdrawal_m[field] += channel_abstraction_cell_m

    if not np.isnan(remaining_irrigation_limit_m3[farmer]):
        remaining_irrigation_limit_m3[farmer] -= channel_abstraction_cell_m3

    channel_abstraction_m3_by_farmer[farmer] += channel_abstraction_cell_m3

    irrigation_water_demand_field_m -= channel_abstraction_cell_m
    irrigation_water_demand_field_m = max(irrigation_water_demand_field_m, 0)

    return irrigation_water_demand_field_m


@njit(cache=True)
def withdraw_reservoir(
    command_area: int,
    field: int,
    farmer: int,
    reservoir_abstraction_m3: npt.NDArray[np.float32],
    available_reservoir_storage_m3: npt.NDArray[np.float32],
    irrigation_water_demand_field_m: np.float32,
    water_withdrawal_m: npt.NDArray[np.float32],
    remaining_irrigation_limit_m3: npt.NDArray[np.float32],
    reservoir_abstraction_m3_by_farmer: npt.NDArray[np.float32],
    maximum_abstraction_reservoir_m3_field: np.float32,
    cell_area: npt.NDArray[np.float32],
) -> np.float32:
    """Withdraw irrigation water from a reservoir for a field.

    Args:
        command_area: Reservoir command-area id for this farmer, or ``-1``.
        field: HRU index.
        farmer: Farmer index.
        reservoir_abstraction_m3: Per-reservoir abstraction accumulator (m³).
        available_reservoir_storage_m3: Available storage per reservoir (m³).
        irrigation_water_demand_field_m: Remaining demand at field (m).
        water_withdrawal_m: Per-field withdrawal accumulator (m).
        remaining_irrigation_limit_m3: Remaining seasonal limit per farmer (m³),
            or ``NaN`` when unlimited.
        reservoir_abstraction_m3_by_farmer: Per-farmer reservoir abstraction accumulator (m³).
        maximum_abstraction_reservoir_m3_field: Max reservoir abstraction allowed for
            this field in current step (m³).
        cell_area: HRU area (m²).

    Returns:
        Updated remaining field demand (m).
    """
    water_demand_cell_m3 = irrigation_water_demand_field_m * cell_area[field]
    assert water_demand_cell_m3 >= 0

    remaining_reservoir_m3 = max(
        available_reservoir_storage_m3[command_area]
        - reservoir_abstraction_m3[command_area],
        0,
    )

    reservoir_abstraction_cell_m3 = min(
        remaining_reservoir_m3,
        water_demand_cell_m3,
        maximum_abstraction_reservoir_m3_field,
    )
    # ensure reservoir abstraction is non-negative
    assert reservoir_abstraction_cell_m3 >= 0

    reservoir_abstraction_m3[command_area] += reservoir_abstraction_cell_m3
    reservoir_abstraction_cell_m = reservoir_abstraction_cell_m3 / cell_area[field]
    water_withdrawal_m[field] += reservoir_abstraction_cell_m

    if not np.isnan(remaining_irrigation_limit_m3[farmer]):
        remaining_irrigation_limit_m3[farmer] -= reservoir_abstraction_cell_m3

    reservoir_abstraction_m3_by_farmer[farmer] += reservoir_abstraction_cell_m3

    irrigation_water_demand_field_m -= reservoir_abstraction_cell_m
    irrigation_water_demand_field_m = np.maximum(
        irrigation_water_demand_field_m, np.float32(0)
    )

    return irrigation_water_demand_field_m


@njit(cache=True)
def withdraw_groundwater(
    farmer: int,
    grid_cell: int,
    field: int,
    groundwater_abstraction_m3: npt.NDArray[np.float32],
    available_groundwater_m3: npt.NDArray[np.float32],
    cell_area: npt.NDArray[np.float32],
    groundwater_depth: npt.NDArray[np.float32],
    well_depth: npt.NDArray[np.float32],
    irrigation_water_demand_field_m: float,
    water_withdrawal_m: npt.NDArray[np.float32],
    remaining_irrigation_limit_m3: npt.NDArray[np.float32],
    groundwater_abstraction_m3_by_farmer: npt.NDArray[np.float32],
    maximum_abstraction_groundwater_m3_field: np.float32,
) -> float:
    """Withdraw irrigation water from groundwater for a field.

    Args:
        farmer: Farmer index.
        grid_cell: Grid-cell index corresponding to field.
        field: HRU index.
        groundwater_abstraction_m3: Per-grid-cell groundwater abstraction accumulator (m³).
        available_groundwater_m3: Available groundwater per grid cell (m³).
        cell_area: HRU area (m²).
        groundwater_depth: Groundwater depth per grid cell (m).
        well_depth: Per-farmer well depth (m).
        irrigation_water_demand_field_m: Remaining demand at field (m).
        water_withdrawal_m: Per-field withdrawal accumulator (m).
        remaining_irrigation_limit_m3: Remaining seasonal limit per farmer (m³),
            or ``NaN`` when unlimited.
        groundwater_abstraction_m3_by_farmer: Per-farmer groundwater abstraction accumulator (m³).
        maximum_abstraction_groundwater_m3_field: Max allowed GW abstraction (m³) for this field.

    Returns:
        Updated remaining field demand (m).
    """
    # groundwater irrigation
    if groundwater_depth[grid_cell] < well_depth[farmer]:
        water_demand_cell_m3 = irrigation_water_demand_field_m * cell_area[field]
        assert water_demand_cell_m3 >= 0

        remaining_groundwater_m3 = max(
            available_groundwater_m3[grid_cell] - groundwater_abstraction_m3[grid_cell],
            0,
        )

        groundwater_abstraction_cell_m3 = min(
            remaining_groundwater_m3,
            water_demand_cell_m3,
            maximum_abstraction_groundwater_m3_field,
        )
        # ensure groundwater abstraction is non-negative
        assert groundwater_abstraction_cell_m3 >= 0

        groundwater_abstraction_m3[grid_cell] += groundwater_abstraction_cell_m3
        groundwater_abstraction_cell_m = (
            groundwater_abstraction_cell_m3 / cell_area[field]
        )
        water_withdrawal_m[field] += groundwater_abstraction_cell_m

        if not np.isnan(remaining_irrigation_limit_m3[farmer]):
            remaining_irrigation_limit_m3[farmer] -= groundwater_abstraction_cell_m3

        groundwater_abstraction_m3_by_farmer[farmer] += groundwater_abstraction_cell_m3

        irrigation_water_demand_field_m -= groundwater_abstraction_cell_m
        irrigation_water_demand_field_m = max(irrigation_water_demand_field_m, 0)

    return irrigation_water_demand_field_m


@njit(cache=True, inline="always")
def get_potential_irrigation_consumption_m(
    topwater: np.float32,
    root_depth_m: np.float32,
    soil_layer_height: ArrayFloat32,
    field_capacity: ArrayFloat32,
    wilting_point: ArrayFloat32,
    w: ArrayFloat32,
    ws: ArrayFloat32,
    saturated_hydraulic_conductivity_m_per_day: ArrayFloat32,
    fraction_irrigated_field: np.float32,
    max_paddy_water_level_farmer: np.float32,
    crop_group: np.float32,  # kept for future dynamic-p logic
    is_paddy: np.bool_,
    minimum_effective_root_depth_m: np.float32,
    depletion_factor: np.float32 = np.float32(0.5),
) -> np.float32:
    """Compute potential irrigation (m) for a field on the current day.

    For paddy: irrigate up to a target ponding level above the current topwater.
    For non-paddy: apply when soil depletion exceeds a fraction of readily
    available water in the root zone, capped by infiltration capacity.

    Args:
        topwater: Current surface water level at the field (m).
        root_depth_m: Current crop root depth (m).
        soil_layer_height: Soil layer heights (m), shape (n_layers,).
        field_capacity: Volumetric water at field capacity per layer, (n_layers,).
        wilting_point: Volumetric water at wilting point per layer, (n_layers,).
        w: Current soil water per layer (same units as FC/WP), (n_layers,).
        ws: Saturation per layer (unused here but kept for signature stability).
        saturated_hydraulic_conductivity_m_per_day: Layer/soil infiltration control (m/day).
        fraction_irrigated_field: Fraction of HRU actually irrigated [0,1].
        max_paddy_water_level_farmer: Target ponding level for paddy (m).
        crop_group: Crop stress group (placeholder for future dynamic p).
        is_paddy: Whether this field is paddy.
        minimum_effective_root_depth_m: Lower bound for effective root depth (m).
        depletion_factor: Multiplier on RAW threshold (default 0.5).

    Returns:
        Potential irrigation depth (m) to apply today (already scaled by
        ``fraction_irrigated_field``).
    """
    assert np.float32(0) <= fraction_irrigated_field <= np.float32(1)

    # Calculate the potential irrigation consumption for the farmer
    if is_paddy:
        # make sure all fields are paddy irrigateds
        # always irrigate to 0.05 m for paddy fields
        potential_irrigation_consumption_m = max(
            max_paddy_water_level_farmer - topwater,
            np.float32(0),
        )
    else:
        # use a minimum root depth of 25 cm, following AQUACROP recommendation
        # see: Reference manual for AquaCrop v7.1 – Chapter 3
        effective_root_depth: np.float32 = np.maximum(
            minimum_effective_root_depth_m, root_depth_m
        )
        root_ratios: npt.NDArray[np.float32] = get_root_ratios(
            effective_root_depth,
            soil_layer_height,
        )

        field_capacity_root_zone = (field_capacity * root_ratios).sum(axis=0)
        wilting_point_root_zone = (wilting_point * root_ratios).sum(axis=0)
        available_water_root_zone = (w * root_ratios).sum(axis=0)

        # calculate the total available soil water for the root zone
        # often refered to as TAW (see AquaCrop v7.1, Chapter 3)
        maximum_available_water = field_capacity_root_zone - wilting_point_root_zone

        # # calculate the depletion coefficient (P_{sto}), which is the fraction of
        # # total available water (TAW) at which stomata start to close
        # p: np.float32 = get_fraction_easily_available_soil_water(
        #     crop_group, potential_evapotranspiration
        # )
        # TODO: Re-enable some dynamic p but not based on PET since this gives
        # issues when used in irrigation
        p = np.float32(0.5)

        readily_available_water_root_zone = maximum_available_water * p

        soil_depletion = field_capacity_root_zone - available_water_root_zone

        if soil_depletion > readily_available_water_root_zone * depletion_factor:
            potential_irrigation_consumption_m = soil_depletion
        else:
            potential_irrigation_consumption_m = np.float32(0)

        infiltration_capacity = saturated_hydraulic_conductivity_m_per_day[0]
        potential_irrigation_consumption_m = np.minimum(
            potential_irrigation_consumption_m, infiltration_capacity
        )

    return potential_irrigation_consumption_m * fraction_irrigated_field


@njit(cache=True)
def get_gross_irrigation_demand_m3(
    day_index: int,
    n: int,
    currently_irrigated_fields: ArrayBool,
    field_indices_by_farmer: TwoDArrayInt32,
    field_indices: ArrayInt32,
    irrigation_efficiency: ArrayFloat32,
    fraction_irrigated_field: ArrayFloat32,
    cell_area: ArrayFloat32,
    crop_map: ArrayInt32,
    topwater: ArrayFloat32,
    root_depth_m: ArrayFloat32,
    soil_layer_height: TwoDArrayFloat32,
    field_capacity: TwoDArrayFloat32,
    wilting_point: TwoDArrayFloat32,
    w: TwoDArrayFloat32,
    ws: TwoDArrayFloat32,
    saturated_hydraulic_conductivity_m_per_day: TwoDArrayFloat32,
    remaining_irrigation_limit_m3_reservoir: ArrayFloat32,
    remaining_irrigation_limit_m3_channel: ArrayFloat32,
    remaining_irrigation_limit_m3_groundwater: ArrayFloat32,
    irrigation_limit_reset_day_index: ArrayInt32,
    cumulative_water_deficit_m3: TwoDArrayFloat32,
    crop_calendar: ThreeDArrayInt32,
    crop_group_numbers: ArrayFloat32,
    paddy_irrigated_crops: ArrayBool,
    current_crop_calendar_rotation_year_index: ArrayInt32,
    max_paddy_water_level: ArrayFloat32,
    minimum_effective_root_depth_m: np.float32,
) -> tuple[ArrayFloat32, ArrayFloat32, ArrayFloat32, ArrayFloat32]:
    """Compute gross irrigation demand per field (m³) and limit-adjusted variants.

    Iterates per farmer and field to compute gross potential irrigation demand
    (m³) from soil/crop state, then applies irrigation-limit corrections for
    each source (in the order reservoir, channel, groundwater) when a limit is present.
    Used to regulate the irrigation behavior of farmers. The farmers are "activated"
    by the given `activation_order`

    Returns:
        tuple[
            np.ndarray,
            np.ndarray,
            np.ndarray,
            np.ndarray,
        ]: A 4-tuple of arrays (all shape = n_fields, dtype float32):
            - gross_potential_irrigation_m3: Unconstrained gross demand (m³) per field.
            - gross_potential_irrigation_m3_limit_adjusted_reservoir: Demand limited by the reservoir cap.
            - gross_potential_irrigation_m3_limit_adjusted_channel: Demand limited by the channel cap.
            - gross_potential_irrigation_m3_limit_adjusted_groundwater: Demand limited by the groundwater cap.
    """
    n_hydrological_response_units: int = cell_area.size
    gross_potential_irrigation_m3: ArrayFloat32 = np.zeros(
        n_hydrological_response_units, dtype=np.float32
    )
    gross_potential_irrigation_m3_limit_adjusted_reservoir: ArrayFloat32 = (
        np.zeros_like(gross_potential_irrigation_m3)
    )
    gross_potential_irrigation_m3_limit_adjusted_channel: ArrayFloat32 = np.zeros_like(
        gross_potential_irrigation_m3
    )
    gross_potential_irrigation_m3_limit_adjusted_groundwater: ArrayFloat32 = (
        np.zeros_like(gross_potential_irrigation_m3)
    )
    # TODO: First part of this function can be parallelized
    for farmer in range(n):
        farmer_fields: npt.NDArray[np.int32] = get_farmer_HRUs(
            field_indices, field_indices_by_farmer, farmer
        )
        irrigation_efficiency_farmer: np.float32 = irrigation_efficiency[farmer]
        for field in farmer_fields:
            if not currently_irrigated_fields[field]:
                continue

            crop: np.int32 = crop_map[field]
            assert crop != -1

            consumption_m: np.float32 = get_potential_irrigation_consumption_m(
                topwater=topwater[field],
                root_depth_m=root_depth_m[field],
                soil_layer_height=soil_layer_height[:, field],
                field_capacity=field_capacity[:, field],
                wilting_point=wilting_point[:, field],
                w=w[:, field],
                ws=ws[:, field],
                saturated_hydraulic_conductivity_m_per_day=saturated_hydraulic_conductivity_m_per_day[
                    :, field
                ],
                fraction_irrigated_field=fraction_irrigated_field[farmer],
                max_paddy_water_level_farmer=max_paddy_water_level[farmer],
                crop_group=crop_group_numbers[crop],
                is_paddy=paddy_irrigated_crops[crop],
                minimum_effective_root_depth_m=minimum_effective_root_depth_m,
            )

            assert consumption_m < 1

            gross_potential_irrigation_m3[field] = (
                consumption_m * cell_area[field]
            ) / irrigation_efficiency_farmer

        farmer_gross_irrigation_demand_m3: np.float32 = gross_potential_irrigation_m3[
            farmer_fields
        ].sum()
        # If the potential irrigation consumption is larger than 0, the farmer needs to abstract water
        # if there is no irrigation limit, no need to adjust the irrigation
        if not np.isnan(remaining_irrigation_limit_m3_reservoir[farmer]):
            if farmer_gross_irrigation_demand_m3 > 0.0:
                irrigation_correction_factor: float = adjust_irrigation_to_limit(
                    farmer=farmer,
                    day_index=day_index,
                    remaining_irrigation_limit_m3=remaining_irrigation_limit_m3_reservoir,
                    cumulative_water_deficit_m3=cumulative_water_deficit_m3,
                    crop_calendar=crop_calendar,
                    crop_rotation_year_index=current_crop_calendar_rotation_year_index,
                    farmer_gross_irrigation_demand_m3=farmer_gross_irrigation_demand_m3,
                    irrigation_efficiency_farmer=irrigation_efficiency_farmer,
                    reset_day_index=irrigation_limit_reset_day_index[farmer],
                )
                assert 0 <= irrigation_correction_factor <= 1

                gross_potential_irrigation_m3_limit_adjusted_reservoir[
                    farmer_fields
                ] = (
                    gross_potential_irrigation_m3[farmer_fields]
                    * irrigation_correction_factor
                )
        else:
            gross_potential_irrigation_m3_limit_adjusted_reservoir[farmer_fields] = (
                gross_potential_irrigation_m3[farmer_fields]
            )

        if not np.isnan(remaining_irrigation_limit_m3_channel[farmer]):
            if farmer_gross_irrigation_demand_m3 > 0.0:
                irrigation_correction_factor: float = adjust_irrigation_to_limit(
                    farmer=farmer,
                    day_index=day_index,
                    remaining_irrigation_limit_m3=remaining_irrigation_limit_m3_channel,
                    cumulative_water_deficit_m3=cumulative_water_deficit_m3,
                    crop_calendar=crop_calendar,
                    crop_rotation_year_index=current_crop_calendar_rotation_year_index,
                    farmer_gross_irrigation_demand_m3=farmer_gross_irrigation_demand_m3,
                    irrigation_efficiency_farmer=irrigation_efficiency_farmer,
                    reset_day_index=irrigation_limit_reset_day_index[farmer],
                )
                assert 0 <= irrigation_correction_factor <= 1

                gross_potential_irrigation_m3_limit_adjusted_channel[farmer_fields] = (
                    gross_potential_irrigation_m3[farmer_fields]
                    * irrigation_correction_factor
                )
        else:
            gross_potential_irrigation_m3_limit_adjusted_channel[farmer_fields] = (
                gross_potential_irrigation_m3[farmer_fields]
            )

        if not np.isnan(remaining_irrigation_limit_m3_groundwater[farmer]):
            if farmer_gross_irrigation_demand_m3 > 0.0:
                irrigation_correction_factor: float = adjust_irrigation_to_limit(
                    farmer=farmer,
                    day_index=day_index,
                    remaining_irrigation_limit_m3=remaining_irrigation_limit_m3_groundwater,
                    cumulative_water_deficit_m3=cumulative_water_deficit_m3,
                    crop_calendar=crop_calendar,
                    crop_rotation_year_index=current_crop_calendar_rotation_year_index,
                    farmer_gross_irrigation_demand_m3=farmer_gross_irrigation_demand_m3,
                    irrigation_efficiency_farmer=irrigation_efficiency_farmer,
                    reset_day_index=irrigation_limit_reset_day_index[farmer],
                )
                assert 0 <= irrigation_correction_factor <= 1

                gross_potential_irrigation_m3_limit_adjusted_groundwater[
                    farmer_fields
                ] = (
                    gross_potential_irrigation_m3[farmer_fields]
                    * irrigation_correction_factor
                )
        else:
            gross_potential_irrigation_m3_limit_adjusted_groundwater[farmer_fields] = (
                gross_potential_irrigation_m3[farmer_fields]
            )

        assert (
            gross_potential_irrigation_m3[farmer_fields] / cell_area[farmer_fields]
        ).max() <= 1

    return (
        gross_potential_irrigation_m3,
        gross_potential_irrigation_m3_limit_adjusted_reservoir,
        gross_potential_irrigation_m3_limit_adjusted_channel,
        gross_potential_irrigation_m3_limit_adjusted_groundwater,
    )


@njit(cache=True)
def abstract_water(
    activation_order: npt.NDArray[np.int32],
    field_indices_by_farmer: npt.NDArray[np.int32],
    field_indices: npt.NDArray[np.int32],
    irrigation_efficiency: npt.NDArray[np.float32],
    surface_irrigated: npt.NDArray[np.bool_],
    well_irrigated: npt.NDArray[np.bool_],
    cell_area: npt.NDArray[np.float32],
    HRU_to_grid: npt.NDArray[np.int32],
    nearest_river_grid_cell: npt.NDArray[np.int32],
    crop_map: npt.NDArray[np.int32],
    available_channel_storage_m3: npt.NDArray[np.float32],
    available_groundwater_m3: npt.NDArray[np.float32],
    groundwater_depth: npt.NDArray[np.float32],
    available_reservoir_storage_m3: npt.NDArray[np.float32],
    command_area_by_farmer: npt.NDArray[np.int32],
    return_fraction: npt.NDArray[np.float32],
    well_depth: npt.NDArray[np.float32],
    remaining_irrigation_limit_m3_reservoir: npt.NDArray[np.float32],
    remaining_irrigation_limit_m3_channel: npt.NDArray[np.float32],
    remaining_irrigation_limit_m3_groundwater: npt.NDArray[np.float32],
    maximum_abstraction_reservoir_m3_by_farmer: npt.NDArray[np.float32],
    maximum_abstraction_channel_m3_by_farmer: npt.NDArray[np.float32],
    maximum_abstraction_groundwater_m3_by_farmer: npt.NDArray[np.float32],
    gross_irrigation_demand_m3_per_field: npt.NDArray[np.float32],
) -> tuple[
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
    npt.NDArray[np.float32],
]:
    """Perform irrigation withdrawals from channel, reservoir, and groundwater.

    Args:
        activation_order: Farmer activation order (e.g., by elevation).
        field_indices_by_farmer: Start/stop indices into ``field_indices`` per farmer.
        field_indices: Flat array of HRU indices for all fields (ordered by farmer).
        irrigation_efficiency: Per-farmer irrigation efficiency.
        surface_irrigated: Whether farmer can access surface-water sources.
        well_irrigated: Whether farmer has a functioning well.
        cell_area: HRU areas.
        HRU_to_grid: Map from HRU to grid cell.
        nearest_river_grid_cell: Nearest river cell per HRU.
        crop_map: Crop id per HRU (``-1`` for none).
        available_channel_storage_m3: Available channel storage per river cell (m³).
        available_groundwater_m3: Available groundwater per grid cell (m³).
        groundwater_depth: Groundwater depth per grid cell (m).
        available_reservoir_storage_m3: Available reservoir storage (m³).
        command_area_by_farmer: Reservoir command area per farmer (``-1`` if none).
        return_fraction: Per-farmer fraction of applied losses that return as return-flow.
        well_depth: Per-farmer well depth (m).
        remaining_irrigation_limit_m3_reservoir: Remaining seasonal limit (m³).
        remaining_irrigation_limit_m3_channel: Remaining seasonal limit (m³).
        remaining_irrigation_limit_m3_groundwater: Remaining seasonal limit (m³).
        maximum_abstraction_reservoir_m3_by_farmer: Cap per farmer (m³) from reservoir.
        maximum_abstraction_channel_m3_by_farmer: Cap per farmer (m³) from channel.
        maximum_abstraction_groundwater_m3_by_farmer: Cap per farmer (m³) from GW.
        gross_irrigation_demand_m3_per_field: Unadjusted gross demand per HRU (m³).

    Returns:
        Tuple with per-source totals and field-level fluxes:
        (channel_by_farmer, reservoir_by_farmer, groundwater_by_farmer,
         water_withdrawal_m, water_consumption_m,
         irrigation_return_flow_m, irrigation_evaporation_m,
         reservoir_abstraction_m3, groundwater_abstraction_m3).
    """
    n_hydrological_response_units = cell_area.size
    water_withdrawal_m = np.zeros(n_hydrological_response_units, dtype=np.float32)
    water_consumption_m = np.zeros(n_hydrological_response_units, dtype=np.float32)

    irrigation_return_flow_m = np.zeros(n_hydrological_response_units, dtype=np.float32)
    irrigation_evaporation_m = np.zeros(n_hydrological_response_units, dtype=np.float32)

    channel_abstraction_m3_by_farmer = np.zeros(activation_order.size, dtype=np.float32)
    reservoir_abstraction_m3_by_farmer = np.zeros(
        activation_order.size, dtype=np.float32
    )
    groundwater_abstraction_m3_by_farmer = np.zeros(
        activation_order.size, dtype=np.float32
    )

    # Because the groundwater and reservoir source are much larger than the surface water
    # and taking out small amounts cannot be represented by
    # floating point numbers, we need to use a separate array that tracks
    # the groundwater abstraction for each field. This is used to
    # then remove the groundwater and reservoir abstraction from the available
    # in one go, reducing the risk of (larger) floating point errors.
    groundwater_abstraction_m3 = np.zeros_like(available_groundwater_m3)
    reservoir_abstraction_m3 = np.zeros_like(available_reservoir_storage_m3)

    for farmer in activation_order:
        farmer_fields = get_farmer_HRUs(field_indices, field_indices_by_farmer, farmer)
        potential_irrigation_consumption_m3_farmer = (
            gross_irrigation_demand_m3_per_field[farmer_fields]
        )

        if potential_irrigation_consumption_m3_farmer.sum() <= 0:
            continue

        command_area_farmer = command_area_by_farmer[farmer]
        return_fraction_farmer = return_fraction[farmer]

        if surface_irrigated[farmer]:
            maximum_abstraction_channel_m3_farmer = (
                maximum_abstraction_channel_m3_by_farmer[farmer]
            )
            maximum_abstraction_channel_m3_by_field = (
                maximum_abstraction_channel_m3_farmer
                * potential_irrigation_consumption_m3_farmer
                / potential_irrigation_consumption_m3_farmer.sum()
            )
            if command_area_farmer == -1:  # -1 means no command area
                pass
            else:
                maximum_abstraction_reservoir_m3_farmer = (
                    maximum_abstraction_reservoir_m3_by_farmer[farmer]
                )
                maximum_abstraction_reservoir_m3_by_field = (
                    maximum_abstraction_reservoir_m3_farmer
                    * potential_irrigation_consumption_m3_farmer
                    / potential_irrigation_consumption_m3_farmer.sum()
                )

        if well_irrigated[farmer]:
            maximum_abstraction_groundwater_m3_farmer = (
                maximum_abstraction_groundwater_m3_by_farmer[farmer]
            )
            maximum_abstraction_groundwater_m3_by_field = (
                maximum_abstraction_groundwater_m3_farmer
                * potential_irrigation_consumption_m3_farmer
                / potential_irrigation_consumption_m3_farmer.sum()
            )

        # loop through all farmers fields and apply irrigation
        for field_index, field in enumerate(farmer_fields):
            grid_cell = HRU_to_grid[field]
            grid_cell_nearest = nearest_river_grid_cell[field]
            if crop_map[field] != -1:
                irrigation_water_demand_field_m = (
                    gross_irrigation_demand_m3_per_field[field] / cell_area[field]
                )
                assert 1 >= irrigation_water_demand_field_m >= 0
                if surface_irrigated[farmer]:
                    # command areas
                    if command_area_farmer != -1:  # -1 means no command area
                        irrigation_water_demand_field_m = withdraw_reservoir(
                            command_area=command_area_farmer,
                            field=field,
                            farmer=farmer,
                            reservoir_abstraction_m3=reservoir_abstraction_m3,
                            available_reservoir_storage_m3=available_reservoir_storage_m3,
                            irrigation_water_demand_field_m=irrigation_water_demand_field_m,
                            water_withdrawal_m=water_withdrawal_m,
                            remaining_irrigation_limit_m3=remaining_irrigation_limit_m3_reservoir,
                            reservoir_abstraction_m3_by_farmer=reservoir_abstraction_m3_by_farmer,
                            maximum_abstraction_reservoir_m3_field=maximum_abstraction_reservoir_m3_by_field[
                                field_index
                            ],
                            cell_area=cell_area,
                        )
                        assert water_withdrawal_m[field] >= 0
                        assert irrigation_water_demand_field_m >= 0

                    irrigation_water_demand_field_m = withdraw_channel(
                        available_channel_storage_m3=available_channel_storage_m3,
                        grid_cell=grid_cell_nearest,
                        cell_area=cell_area,
                        field=field,
                        farmer=farmer,
                        water_withdrawal_m=water_withdrawal_m,
                        irrigation_water_demand_field_m=irrigation_water_demand_field_m,
                        remaining_irrigation_limit_m3=remaining_irrigation_limit_m3_channel,
                        channel_abstraction_m3_by_farmer=channel_abstraction_m3_by_farmer,
                        maximum_abstraction_channel_m3_field=maximum_abstraction_channel_m3_by_field[
                            field_index
                        ],
                        minimum_channel_storage_m3=100.0,
                    )
                    assert water_withdrawal_m[field] >= 0
                    assert irrigation_water_demand_field_m >= 0

                if well_irrigated[farmer]:
                    irrigation_water_demand_field_m = withdraw_groundwater(
                        farmer=farmer,
                        field=field,
                        grid_cell=grid_cell,
                        groundwater_abstraction_m3=groundwater_abstraction_m3,
                        available_groundwater_m3=available_groundwater_m3,
                        cell_area=cell_area,
                        groundwater_depth=groundwater_depth,
                        well_depth=well_depth,
                        irrigation_water_demand_field_m=irrigation_water_demand_field_m,
                        water_withdrawal_m=water_withdrawal_m,
                        remaining_irrigation_limit_m3=remaining_irrigation_limit_m3_groundwater,
                        groundwater_abstraction_m3_by_farmer=groundwater_abstraction_m3_by_farmer,
                        maximum_abstraction_groundwater_m3_field=maximum_abstraction_groundwater_m3_by_field[
                            field_index
                        ],
                    )
                    assert irrigation_water_demand_field_m >= 0
                    assert water_withdrawal_m[field] >= 0

                assert (
                    irrigation_water_demand_field_m >= -1e15
                )  # Make sure irrigation water demand is zero, or positive. Allow very small error.

                assert water_consumption_m[field] >= 0
                assert water_withdrawal_m[field] >= 0
                assert 1 >= return_fraction_farmer >= 0

                water_consumption_m[field] = (
                    water_withdrawal_m[field] * irrigation_efficiency[farmer]
                )
                irrigation_loss_m = (
                    water_withdrawal_m[field] - water_consumption_m[field]
                )
                assert irrigation_loss_m >= 0
                irrigation_return_flow_m[field] = (
                    irrigation_loss_m * return_fraction_farmer
                )
                irrigation_evaporation_m[field] = (
                    irrigation_loss_m - irrigation_return_flow_m[field]
                )

    return (
        channel_abstraction_m3_by_farmer,
        reservoir_abstraction_m3_by_farmer,
        groundwater_abstraction_m3_by_farmer,
        water_withdrawal_m,
        water_consumption_m,
        irrigation_return_flow_m,
        irrigation_evaporation_m,
        reservoir_abstraction_m3,
        groundwater_abstraction_m3,
    )


@njit(cache=True)
def plant(
    n: int,
    day_index: int,
    crop_calendar: np.ndarray,
    current_crop_calendar_rotation_year_index: np.ndarray,
    crop_map: np.ndarray,
    crop_harvest_age_days: np.ndarray,
    cultivation_cost: np.ndarray | int | float,
    region_ids_per_farmer: np.ndarray,
    field_indices_by_farmer: np.ndarray,
    field_indices: np.ndarray,
    field_size_per_farmer: np.ndarray,
    all_loans_annual_cost: np.ndarray,
    loan_tracker: np.ndarray,
    interest_rate: np.ndarray,
    farmers_going_out_of_business: bool,
) -> tuple[npt.NDArray[np.int32], npt.NDArray[np.int64]]:
    """Determines when and what crop should be planted, by comparing the current day to the next plant day. Also sets the harvest age of the plant.

    Args:
        n: Number of farmers.
        day_index: Current day index (0-indexed; Jan 1 == 0).
        crop_calendar: Crop calendar for each farmer. Each row is a farmer, and each column is a crop.
            Each crop is a list of [crop_type, planting_day, growing_days, crop_year_index].
            Planting day is 0-indexed (Jan 1 == 0).
            Growing days is the number of days the crop grows.
            Crop year index is the index of the year in the crop rotation.
        current_crop_calendar_rotation_year_index: Current crop calendar rotation year index for each farmer.
        crop_map: Map of the currently growing crops.
        crop_harvest_age_days: Array that contains the harvest age of each field in days.
        cultivation_cost: Cultivation cost per farmer per crop in m2. If this is a single value, it is used for all farmers and crops.
        region_ids_per_farmer: Region IDs for each farmer.
            This is used to determine the cultivation cost for each farmer.
        field_indices_by_farmer: This array contains the indices where the fields of a farmer are stored in `field_indices`.
        field_indices: This array contains the indices of all fields, ordered by farmer. In other words, if a farmer owns multiple fields, the indices of the fields are indices.
        field_size_per_farmer: Field size per farmer in m2
        all_loans_annual_cost: Annual cost of all loans for each farmer and crop.
            This is a 3D array with shape (n, 1, 4), where n is the number of farmers.
            The first dimension is the farmer, the second dimension is the crop, and the third dimension is the loan type.
        loan_tracker: Loan tracker for each farmer and crop. This is a 3D array with shape (n, 1, 4).
        interest_rate: Interest rate for each farmer.
        farmers_going_out_of_business: If True, farmers are going out of business. Not implemented yet.

    Returns:
        plant: Subarray map of what crops are planted this day.
        farmers_selling_land: Indices of farmers selling land (currently always empty).
    """
    assert farmers_going_out_of_business is False, (
        "Farmers going out of business not implemented."
    )

    plant = np.full_like(crop_map, -1, dtype=np.int32)
    sell_land = np.zeros(n, dtype=np.bool_)

    planting_farmers_per_season = (crop_calendar[:, :, 1] == day_index) & (
        crop_calendar[:, :, 3]
        == current_crop_calendar_rotation_year_index[:, np.newaxis]
    )
    planting_farmers = planting_farmers_per_season.sum(axis=1)

    assert planting_farmers.max() <= 1, "Multiple crops planted on the same day"

    planting_farmers_idx = np.where(planting_farmers == 1)[0]
    if not planting_farmers_idx.size == 0:
        crop_rotation = np.argmax(
            planting_farmers_per_season[planting_farmers_idx], axis=1
        )

        assert planting_farmers_idx.size == crop_rotation.size

        for i in range(planting_farmers_idx.size):
            farmer_idx = planting_farmers_idx[i]
            farmer_crop_rotation = crop_rotation[i]

            farmer_fields = get_farmer_HRUs(
                field_indices, field_indices_by_farmer, farmer_idx
            )
            farmer_crop_data = crop_calendar[farmer_idx, farmer_crop_rotation]
            farmer_crop = farmer_crop_data[0]
            field_harvest_age = farmer_crop_data[2]

            assert farmer_crop != -1

            if isinstance(cultivation_cost, (int, float)):
                cultivation_cost_farmer = cultivation_cost
            else:
                farmer_region_id = region_ids_per_farmer[farmer_idx]
                cultivation_cost_farmer = (
                    cultivation_cost[farmer_region_id, farmer_crop]
                    * field_size_per_farmer[farmer_idx]
                )
            assert not np.isnan(cultivation_cost_farmer)

            interest_rate_farmer = interest_rate[farmer_idx]
            loan_duration = 1
            annual_cost_input_loan = cultivation_cost_farmer * (
                interest_rate_farmer
                * (1 + interest_rate_farmer) ** loan_duration
                / ((1 + interest_rate_farmer) ** loan_duration - 1)
            )
            for i in range(4):
                if all_loans_annual_cost[farmer_idx, 0, i] == 0:
                    all_loans_annual_cost[farmer_idx, 0, i] += (
                        annual_cost_input_loan  # Add the amount to the input specific loan
                    )
                    loan_tracker[farmer_idx, 0, i] = loan_duration
                    break  # Exit the loop after adding to the first zero value

            all_loans_annual_cost[farmer_idx, -1, 0] += (
                annual_cost_input_loan  # Add the amount to the total loan amount
            )

            for field in farmer_fields:
                # a crop is still growing here.
                if crop_harvest_age_days[field] != -1:
                    continue
                plant[field] = farmer_crop
                crop_harvest_age_days[field] = field_harvest_age

    farmers_selling_land = np.where(sell_land)[0]
    return plant, farmers_selling_land


@njit(cache=True)
def arrays_equal(a: npt.NDArray, b: npt.NDArray) -> bool:
    """Check element-wise equality for two same-shaped arrays.

    Args:
        a: First array.
        b: Second array.

    Returns:
        True if every element is equal, otherwise False.
    """
    for i in range(a.size):
        if a.flat[i] != b.flat[i]:
            return True if a.flat[i] != b.flat[i] else False
    return True


@njit(cache=True)
def find_matching_rows(
    arr: npt.NDArray[np.int32], target_row: npt.NDArray[np.int32]
) -> npt.NDArray[np.bool_]:
    """Find rows in ``arr`` that match ``target_row`` exactly.

    Args:
        arr: 2D array of shape (n_rows, n_cols).
        target_row: 1D array of shape (n_cols,).

    Returns:
        Boolean mask of length ``n_rows`` where True indicates a match.
    """
    n_rows = arr.shape[0]
    matches = np.empty(n_rows, dtype=np.bool_)
    for i in range(n_rows):
        matches[i] = True
        for j in range(arr.shape[1]):
            if arr[i, j] != target_row[j]:
                matches[i] = False
                break
    return matches


@njit(cache=True)
def find_most_similar_index(
    target_series: npt.NDArray[np.float32],
    yield_ratios: npt.NDArray[np.float32],
    groups: npt.NDArray[np.bool_],
) -> int:
    """Find index (within ``groups``) whose series is closest to ``target_series``.

    Args:
        target_series: 1D reference series.
        yield_ratios: 2D matrix where each row is a candidate series.
        groups: Boolean mask selecting allowable rows.

    Returns:
        Index (into ``yield_ratios``) of the closest series among ``groups``.
    """
    n = groups.size
    indices = []
    for i in range(n):
        if groups[i]:
            indices.append(i)
    n_indices = len(indices)
    distances = np.empty(n_indices, dtype=yield_ratios.dtype)
    for idx in range(n_indices):
        i = indices[idx]
        diff = yield_ratios[i] - target_series
        distances[idx] = np.linalg.norm(diff)
    min_idx = 0
    min_dist = distances[0]
    for idx in range(1, n_indices):
        if distances[idx] < min_dist:
            min_dist = distances[idx]
            min_idx = idx
    return indices[min_idx]


@njit(cache=True, parallel=True)
def crop_profit_difference_njit_parallel(
    yearly_profits: npt.NDArray[np.float32],
    crop_elevation_group: npt.NDArray[np.int32],
    unique_crop_groups: npt.NDArray[np.int32],
    group_indices: npt.NDArray[np.int32],
    crop_calendar: npt.NDArray[np.int32],
    unique_crop_calendars: npt.NDArray[np.int32],
    p_droughts: npt.NDArray[np.float32],
    past_window: int,
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.int32]]:
    """Compute profit gains for alternative crop calendars (parallel over groups).

    Args:
        yearly_profits: Per-farmer yearly profits (recent years in columns).
        crop_elevation_group: Per-farmer feature rows for grouping.
        unique_crop_groups: Unique group keys.
        group_indices: For each farmer, index into ``unique_crop_groups``.
        crop_calendar: Per-farmer crop calendar.
        unique_crop_calendars: Candidate alternative calendars (rows).
        p_droughts: Drought probabilities (not used directly in this body).
        past_window: Number of past years to use (with a decaying weight).

    Returns:
        Tuple of:
            - gains_adaptation: Profit gain per farmer for each candidate calendar.
            - new_farmer_id: Example “nearest” farmer id per candidate and farmer.
    """
    n_groups: int = len(unique_crop_groups)
    n_calendars: int = len(unique_crop_calendars)
    n_rotation: int = unique_crop_calendars.shape[1]

    unique_profit_gain = np.full((n_groups, n_calendars), 0.0, dtype=np.float32)
    id_to_switch_to = np.full((n_groups, n_calendars), -1, dtype=np.int32)

    # Precompute weights
    weights = np.empty(past_window, dtype=yearly_profits.dtype)
    weight_total = 0.0
    if past_window > 1:
        weight_step = (0.5 - 0.2) / (past_window - 1)
    else:
        weight_step = 0.0

    for j in range(past_window):
        w = 0.5 - j * weight_step
        weights[j] = w
        weight_total += w

    # Parallel only over groups
    for group_id in prange(n_groups):  # ty: ignore[not-iterable]
        unique_group = unique_crop_groups[group_id]
        unique_farmer_groups = find_matching_rows(crop_elevation_group, unique_group)

        # build mask & collect candidate calendar‐indices
        candidate_idxs = np.empty(n_calendars, dtype=np.int32)
        num_cand = 0
        for i in range(n_calendars):
            match = True
            for j in range(n_rotation):
                if unique_crop_calendars[i, j] != unique_group[j]:
                    match = False
                    break
            if not match:
                candidate_idxs[num_cand] = i
                num_cand += 1

        metadata = unique_group[n_rotation:]

        # sequentially over each candidate (same order as original)
        for c in range(num_cand):
            cal_idx = candidate_idxs[c]
            unique_rotation = unique_crop_calendars[cal_idx]

            # form the “other” group key
            other_key = np.empty(unique_group.shape[0], unique_group.dtype)
            for k in range(n_rotation):
                other_key[k] = unique_rotation[k]
            for k in range(n_rotation, other_key.shape[0]):
                other_key[k] = metadata[k - n_rotation]

            unique_farmer_other = find_matching_rows(crop_elevation_group, other_key)

            if not np.any(unique_farmer_other):
                continue

            # current group weighted average
            current_sum = 0.0
            count_cur = 0
            for i in range(unique_farmer_groups.size):
                if unique_farmer_groups[i]:
                    wp = 0.0
                    for j in range(past_window):
                        wp += yearly_profits[i, j] * weights[j]
                    current_sum += wp / weight_total
                    count_cur += 1
            current_avg = current_sum / count_cur if count_cur > 0 else 0.0

            # candidate group weighted average
            cand_sum = 0.0
            count_cand = 0
            for i in range(unique_farmer_other.size):
                if unique_farmer_other[i]:
                    wp = 0.0
                    for j in range(past_window):
                        wp += yearly_profits[i, j] * weights[j]
                    cand_sum += wp / weight_total
                    count_cand += 1
            cand_avg = cand_sum / count_cand if count_cand > 0 else 0.0

            gain = cand_avg - current_avg
            if gain != gain:  # NaN check
                gain = 0.0
            else:
                id_to_switch_to[group_id, c] = find_most_similar_index(
                    gain, yearly_profits, unique_farmer_other
                )

            unique_profit_gain[group_id, c] = gain

    # Re‐index per‐farmer
    gains_adaptation = unique_profit_gain[group_indices, :]
    new_farmer_id = id_to_switch_to[group_indices, :]

    return gains_adaptation, new_farmer_id


@njit
def gev_ppf_scalar(u: float, c: float, loc: float, scale: float) -> float:
    """Scalar GEV inverse CDF.

    CDF = Cumulative Distribution Function
    PPF = Percent Point Function = Inverse CDF
    GEV = Generalized Extreme Value distribution

    Args:
        u: A uniform random variable in (0,1).
        c: GEV shape parameter.
        loc: GEV location parameter.
        scale: GEV scale parameter.

    Returns:
        The inverse CDF value for a given uniform random variable u and GEV parameters c, loc, scale.

    """
    if c != 0.0:
        return loc + scale * ((-np.log(u)) ** (-c) - 1.0) / c
    else:
        return loc - scale * np.log(-np.log(u))


@njit(cache=True, parallel=True)
def compute_premiums_and_best_contracts_numba(
    gev_params: npt.NDArray[np.float64],
    values_history: npt.NDArray[np.float64],
    losses: npt.NDArray[np.float64],
    strike_vals: npt.NDArray[np.float64],
    exit_vals: npt.NDArray[np.float64],
    rate_vals: npt.NDArray[np.float64],
    n_sims: int,
    seed: int = 42,
) -> tuple[
    npt.NDArray[np.int64],
    npt.NDArray[np.int64],
    npt.NDArray[np.int64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """Computes the best insurance contracts for each agent based on GEV parameters, historical SPEI data, and losses.

    For each agent, loop once over (strike, exit):
    1) Monte Carlo → expected_ratio
    2) Historical SPEI → sum_ratio_sq, sum_ratio_loss )
    3) Over all rates → compute both premium = rate * expected_ratio
                            and RMSE via quadratic form,
        tracking the (strike_idx, exit_idx, rate_idx) that minimizes RMSE.

    Returns:
        best_strike_idx : (n_agents,)  index into strike_vals
        best_exit_idx   : (n_agents,)  index into exit_vals
        best_rate_idx   : (n_agents,)  index into rate_vals
        best_rmse       : (n_agents,)  the minimal RMSE for each agent
        best_premium    : (n_agents,)  the premium corresponding to that minimal‐RMSE contract
    """
    np.random.seed(seed)
    n_agents = gev_params.shape[0]
    n_years = values_history.shape[1]
    n_strikes = strike_vals.shape[0]
    n_exits = exit_vals.shape[0]
    n_rates = rate_vals.shape[0]

    # Output arrays
    best_strike_idx = np.empty(n_agents, dtype=np.int64)
    best_exit_idx = np.empty(n_agents, dtype=np.int64)
    best_rate_idx = np.empty(n_agents, dtype=np.int64)
    best_rmse_arr = np.empty(n_agents, dtype=np.float64)
    best_prem_arr = np.empty(n_agents, dtype=np.float64)

    for agent_idx in prange(n_agents):  # ty: ignore[not-iterable]
        shape = -gev_params[agent_idx, 0]
        loc = gev_params[agent_idx, 1]
        scale = gev_params[agent_idx, 2]

        # Precompute sum of squared losses for this agent
        loss_sq_sum = 0.0
        for yr in range(n_years):
            loss_sq_sum += losses[agent_idx, yr] ** 2

        # Initialize "best so far" placeholders
        best_rmse_val = 1e20
        best_s_idx = 0
        best_e_idx = 0
        best_r_idx = 0
        best_premium = 0.0

        # Loop once over all (strike, exit) pairs
        for strike_idx in range(n_strikes):
            strike = strike_vals[strike_idx]

            for exit_idx in range(n_exits):
                exit_threshold = exit_vals[exit_idx]
                if exit_threshold >= strike:
                    continue
                denom = strike - exit_threshold

                # Monte Carlo → expected_ratio
                expected_ratio = 0.0
                for _ in range(n_sims):
                    u = np.random.random()  # Uniform(0,1)
                    spei_val = gev_ppf_scalar(u, shape, loc, scale)
                    shortfall = strike - spei_val
                    if shortfall <= 0.0:
                        continue
                    if shortfall > denom:
                        shortfall = denom
                    expected_ratio += shortfall / denom
                expected_ratio /= n_sims

                sum_ratio_sq = 0.0
                sum_ratio_loss = 0.0
                for yr in range(n_years):
                    shortfall = strike - values_history[agent_idx, yr]
                    if shortfall <= 0.0:
                        ratio = 0.0
                    elif shortfall >= denom:
                        ratio = 1.0
                    else:
                        ratio = shortfall / denom

                    sum_ratio_sq += ratio * ratio
                    sum_ratio_loss += ratio * losses[agent_idx, yr]

                # Over all candidate rates, compute premium & RMSE via quadratic form
                for rate_idx in range(n_rates):
                    rate = rate_vals[rate_idx]
                    # RMSE’s sum of squared errors:
                    sse = (
                        (rate * rate) * sum_ratio_sq
                        - 2.0 * rate * sum_ratio_loss
                        + loss_sq_sum
                    )
                    rmse_val = np.sqrt(sse / n_years)

                    if rmse_val < best_rmse_val:
                        best_rmse_val = rmse_val
                        best_s_idx = strike_idx
                        best_e_idx = exit_idx
                        best_r_idx = rate_idx
                        best_premium = rate * expected_ratio

        # Store best‐so‐far for this agent
        best_strike_idx[agent_idx] = best_s_idx
        best_exit_idx[agent_idx] = best_e_idx
        best_rate_idx[agent_idx] = best_r_idx
        best_rmse_arr[agent_idx] = best_rmse_val
        best_prem_arr[agent_idx] = best_premium

    return (
        best_strike_idx,
        best_exit_idx,
        best_rate_idx,
        best_rmse_arr,
        best_prem_arr,
    )



"""File adapter for FAO datasets."""

from __future__ import annotations

import zipfile
from typing import Any

import numpy as np
import pandas as pd
import xarray as xr

from geb.workflows.io import fetch_and_save

from ..workflows.conversions import (
    M49_to_ISO3,
)
from .base import Adapter


class GMIA(Adapter):
    """Adapter for Global Map of Irrigation Areas datasets."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the Global Map of Irrigation Areas adapter."""
        super().__init__(*args, **kwargs)

    def fetch(self, url: str) -> GMIA:
        """Fetch the dataset from the given URL if not already present.

        Args:
            url: The URL to fetch the dataset from.

        Returns:
            The GMIA adapter instance.
        """
        if not self.is_ready:
            download_path = self.path.with_suffix(".zip")
            fetch_and_save(url=url, file_path=download_path)

            # unzip the file
            with zipfile.ZipFile(download_path, "r") as zip_ref:
                files = zip_ref.namelist()
                assert len(files) == 1, "Expected exactly one file in the zip."
                file = files[0]
                zip_ref.extract(file, self.root)

            extracted_file_path = self.root / file
            extracted_file_path.rename(self.path)  # rename to desired filename
            download_path.unlink()  # Remove the downloaded zip file
        return self

    def read(self, **kwargs: Any) -> xr.DataArray:
        """Read the dataset into an xarray DataArray.

        Also sets the appropriate CRS.

        Args:
            **kwargs: Additional keyword arguments to pass to the reader function.

        Returns:
            The dataset as an xarray DataArray with appropriate CRS.
        """
        da = super().read(**kwargs)
        da = da.sel(band=1)
        da = da.rio.write_crs("EPSG:4326")
        return da


class FAOSTAT(Adapter):
    """Adapter for FAOSTAT datasets."""

    def fetch(self, url: str) -> FAOSTAT:
        """Fetch the dataset from the given URL if not already present.

        Args:
            url: The URL to fetch the dataset from.

        Returns:
            The FAOSTAT adapter instance.
        """
        if not self.is_ready:
            download_path = self.path.with_suffix(".zip")
            fetch_and_save(url=url, file_path=download_path, show_progress=True)

            # unzip the file
            with zipfile.ZipFile(download_path, "r") as zip_ref:
                file = "Prices_E_All_Data.csv"
                zip_ref.extract(file, self.root)

            extracted_file_path = self.root / file

            # Read and filter the CSV
            df = pd.read_csv(extracted_file_path, encoding="ISO-8859-1")

            # Filter for Producer Price (USD/tonne)
            df = df[df["Element"] == "Producer Price (USD/tonne)"]

            # Save as parquet
            df.to_parquet(self.path, index=False)

            # Cleanup
            extracted_file_path.unlink()
            download_path.unlink()

        return self

    def read(self) -> pd.DataFrame:
        """Read the dataset.

        Returns:
            The dataset as a pandas DataFrame.
        """
        df: pd.DataFrame = pd.read_parquet(self.path)

        year_columns: list[str] = [
            col for col in df.columns if col.startswith("Y") and col[1:].isdigit()
        ]
        df = df[["Area Code (M49)", "Item", *year_columns]]

        df = df.rename(columns={col: int(col[1:]) for col in year_columns})
        year_columns: list[int] = [int(col[1:]) for col in year_columns]

        # M49 codes are FAO's area codes
        df["Area Code (M49)"] = df["Area Code (M49)"].apply(
            lambda x: int(x[1:]) if x.startswith("'") else int(x)
        )

        # drop regions that do not have ISO3 codes
        df = df[~df["Area Code (M49)"].isin([58, 200, 230, 891, 736])]
        df["ISO3"] = df["Area Code (M49)"].map(M49_to_ISO3)
        assert df["ISO3"].isnull().sum() == 0, (
            "Some M49 codes could not be mapped to ISO3."
        )
        df = df.drop(columns=["Area Code (M49)"])

        for year in year_columns:
            df[year] = df[year] / 1000.0  # Convert from USD/tonne to USD/kg
            df[year] = df[year].astype(np.float64)

        df = df.rename(
            columns={
                "Item": "crop",
            }
        )

        df["crop"] = df["crop"].str.lower()

        return df



"""Utilities to download MERIT Hydro tiles for a given bounding box.

This module provides a streaming downloader that extracts only the needed
GeoTIFF tiles from the remote 30x30-degree tar packages hosted by MERIT Hydro.
It avoids saving the full tar archives to disk by iterating over the HTTP
response stream and writing only the requested members.

Notes:
    - MERIT Hydro distributes 5x5-degree tiles grouped in 30x30-degree tar files.
      Tile filenames follow the pattern "n30w120_elv.tif" where the lat/lon refer
      to the lower-left corner of the tile in degrees. Package tar files follow
      the pattern "elv_n30w120.tar" where the coordinates indicate the lower-left
      corner of the 30x30 group. See the MERIT Hydro documentation for details.
    - Coverage spans from S60 to N90. Some packages or tiles are fully ocean and
      therefore do not exist. This module distinguishes between "missing because
      ocean/not provided" and actual download errors.

"""

from __future__ import annotations

import os
import tarfile
import time
from pathlib import Path
from typing import IO, Any, Iterable

import numpy as np
import requests
import rioxarray as rxr
import xarray as xr
from requests.auth import HTTPBasicAuth
from rioxarray import merge
from tqdm import tqdm

from geb.workflows.raster import convert_nodata

from .base import Adapter

# MERIT Hydro is only available over land, so not all tiles exist. This is a set of
# all available 5x5-degree tile names which we use to check if the tile
# should exist before attempting to read and if that fails download it.
available_tiles: set[str] = {
    "n00e005",
    "n00e010",
    "n00e015",
    "n00e020",
    "n00e025",
    "n00e030",
    "n00e035",
    "n00e040",
    "n00e045",
    "n00e070",
    "n00e095",
    "n00e100",
    "n00e105",
    "n00e110",
    "n00e115",
    "n00e120",
    "n00e125",
    "n00e130",
    "n00e150",
    "n00e165",
    "n00e170",
    "n00w005",
    "n00w010",
    "n00w050",
    "n00w055",
    "n00w060",
    "n00w065",
    "n00w070",
    "n00w075",
    "n00w080",
    "n00w085",
    "n00w090",
    "n00w095",
    "n00w160",
    "n00w165",
    "n00w180",
    "n05e000",
    "n05e005",
    "n05e010",
    "n05e015",
    "n05e020",
    "n05e025",
    "n05e030",
    "n05e035",
    "n05e040",
    "n05e045",
    "n05e050",
    "n05e070",
    "n05e075",
    "n05e080",
    "n05e090",
    "n05e095",
    "n05e100",
    "n05e105",
    "n05e110",
    "n05e115",
    "n05e120",
    "n05e125",
    "n05e130",
    "n05e135",
    "n05e140",
    "n05e145",
    "n05e150",
    "n05e155",
    "n05e160",
    "n05e165",
    "n05e170",
    "n05w005",
    "n05w010",
    "n05w015",
    "n05w055",
    "n05w060",
    "n05w065",
    "n05w070",
    "n05w075",
    "n05w080",
    "n05w085",
    "n05w090",
    "n05w165",
    "n10e000",
    "n10e005",
    "n10e010",
    "n10e015",
    "n10e020",
    "n10e025",
    "n10e030",
    "n10e035",
    "n10e040",
    "n10e045",
    "n10e050",
    "n10e070",
    "n10e075",
    "n10e080",
    "n10e090",
    "n10e095",
    "n10e100",
    "n10e105",
    "n10e110",
    "n10e115",
    "n10e120",
    "n10e125",
    "n10e135",
    "n10e140",
    "n10e145",
    "n10e160",
    "n10e165",
    "n10e170",
    "n10w005",
    "n10w010",
    "n10w015",
    "n10w020",
    "n10w025",
    "n10w060",
    "n10w065",
    "n10w070",
    "n10w075",
    "n10w080",
    "n10w085",
    "n10w090",
    "n10w095",
    "n10w110",
    "n15e000",
    "n15e005",
    "n15e010",
    "n15e015",
    "n15e020",
    "n15e025",
    "n15e030",
    "n15e035",
    "n15e040",
    "n15e045",
    "n15e050",
    "n15e055",
    "n15e070",
    "n15e075",
    "n15e080",
    "n15e085",
    "n15e090",
    "n15e095",
    "n15e100",
    "n15e105",
    "n15e110",
    "n15e115",
    "n15e120",
    "n15e145",
    "n15e165",
    "n15w005",
    "n15w010",
    "n15w015",
    "n15w020",
    "n15w025",
    "n15w030",
    "n15w065",
    "n15w070",
    "n15w075",
    "n15w080",
    "n15w085",
    "n15w090",
    "n15w095",
    "n15w100",
    "n15w105",
    "n15w110",
    "n15w115",
    "n15w155",
    "n15w160",
    "n15w170",
    "n20e000",
    "n20e005",
    "n20e010",
    "n20e015",
    "n20e020",
    "n20e025",
    "n20e030",
    "n20e035",
    "n20e040",
    "n20e045",
    "n20e050",
    "n20e055",
    "n20e065",
    "n20e070",
    "n20e075",
    "n20e080",
    "n20e085",
    "n20e090",
    "n20e095",
    "n20e100",
    "n20e105",
    "n20e110",
    "n20e115",
    "n20e120",
    "n20e125",
    "n20e130",
    "n20e135",
    "n20e140",
    "n20e145",
    "n20e150",
    "n20w005",
    "n20w010",
    "n20w015",
    "n20w020",
    "n20w075",
    "n20w080",
    "n20w085",
    "n20w090",
    "n20w095",
    "n20w100",
    "n20w105",
    "n20w110",
    "n20w115",
    "n20w120",
    "n20w160",
    "n20w165",
    "n20w170",
    "n25e000",
    "n25e005",
    "n25e010",
    "n25e015",
    "n25e020",
    "n25e025",
    "n25e030",
    "n25e035",
    "n25e040",
    "n25e045",
    "n25e050",
    "n25e055",
    "n25e060",
    "n25e065",
    "n25e070",
    "n25e075",
    "n25e080",
    "n25e085",
    "n25e090",
    "n25e095",
    "n25e100",
    "n25e105",
    "n25e110",
    "n25e115",
    "n25e120",
    "n25e125",
    "n25e130",
    "n25e140",
    "n25w005",
    "n25w010",
    "n25w015",
    "n25w020",
    "n25w080",
    "n25w085",
    "n25w090",
    "n25w095",
    "n25w100",
    "n25w105",
    "n25w110",
    "n25w115",
    "n25w120",
    "n25w175",
    "n25w180",
    "n30e000",
    "n30e005",
    "n30e010",
    "n30e015",
    "n30e020",
    "n30e025",
    "n30e030",
    "n30e035",
    "n30e040",
    "n30e045",
    "n30e050",
    "n30e055",
    "n30e060",
    "n30e065",
    "n30e070",
    "n30e075",
    "n30e080",
    "n30e085",
    "n30e090",
    "n30e095",
    "n30e100",
    "n30e105",
    "n30e110",
    "n30e115",
    "n30e120",
    "n30e125",
    "n30e130",
    "n30e135",
    "n30e140",
    "n30w005",
    "n30w010",
    "n30w020",
    "n30w065",
    "n30w080",
    "n30w085",
    "n30w090",
    "n30w095",
    "n30w100",
    "n30w105",
    "n30w110",
    "n30w115",
    "n30w120",
    "n30w125",
    "n35e000",
    "n35e005",
    "n35e010",
    "n35e015",
    "n35e020",
    "n35e025",
    "n35e030",
    "n35e035",
    "n35e040",
    "n35e045",
    "n35e050",
    "n35e055",
    "n35e060",
    "n35e065",
    "n35e070",
    "n35e075",
    "n35e080",
    "n35e085",
    "n35e090",
    "n35e095",
    "n35e100",
    "n35e105",
    "n35e110",
    "n35e115",
    "n35e120",
    "n35e125",
    "n35e130",
    "n35e135",
    "n35e140",
    "n35w005",
    "n35w010",
    "n35w025",
    "n35w030",
    "n35w035",
    "n35w075",
    "n35w080",
    "n35w085",
    "n35w090",
    "n35w095",
    "n35w100",
    "n35w105",
    "n35w110",
    "n35w115",
    "n35w120",
    "n35w125",
    "n40e000",
    "n40e005",
    "n40e010",
    "n40e015",
    "n40e020",
    "n40e025",
    "n40e030",
    "n40e035",
    "n40e040",
    "n40e045",
    "n40e050",
    "n40e055",
    "n40e060",
    "n40e065",
    "n40e070",
    "n40e075",
    "n40e080",
    "n40e085",
    "n40e090",
    "n40e095",
    "n40e100",
    "n40e105",
    "n40e110",
    "n40e115",
    "n40e120",
    "n40e125",
    "n40e130",
    "n40e135",
    "n40e140",
    "n40e145",
    "n40w005",
    "n40w010",
    "n40w060",
    "n40w065",
    "n40w070",
    "n40w075",
    "n40w080",
    "n40w085",
    "n40w090",
    "n40w095",
    "n40w100",
    "n40w105",
    "n40w110",
    "n40w115",
    "n40w120",
    "n40w125",
    "n45e000",
    "n45e005",
    "n45e010",
    "n45e015",
    "n45e020",
    "n45e025",
    "n45e030",
    "n45e035",
    "n45e040",
    "n45e045",
    "n45e050",
    "n45e055",
    "n45e060",
    "n45e065",
    "n45e070",
    "n45e075",
    "n45e080",
    "n45e085",
    "n45e090",
    "n45e095",
    "n45e100",
    "n45e105",
    "n45e110",
    "n45e115",
    "n45e120",
    "n45e125",
    "n45e130",
    "n45e135",
    "n45e140",
    "n45e145",
    "n45e150",
    "n45e155",
    "n45w005",
    "n45w010",
    "n45w055",
    "n45w060",
    "n45w065",
    "n45w070",
    "n45w075",
    "n45w080",
    "n45w085",
    "n45w090",
    "n45w095",
    "n45w100",
    "n45w105",
    "n45w110",
    "n45w115",
    "n45w120",
    "n45w125",
    "n45w130",
    "n50e000",
    "n50e005",
    "n50e010",
    "n50e015",
    "n50e020",
    "n50e025",
    "n50e030",
    "n50e035",
    "n50e040",
    "n50e045",
    "n50e050",
    "n50e055",
    "n50e060",
    "n50e065",
    "n50e070",
    "n50e075",
    "n50e080",
    "n50e085",
    "n50e090",
    "n50e095",
    "n50e100",
    "n50e105",
    "n50e110",
    "n50e115",
    "n50e120",
    "n50e125",
    "n50e130",
    "n50e135",
    "n50e140",
    "n50e150",
    "n50e155",
    "n50e160",
    "n50e165",
    "n50e170",
    "n50e175",
    "n50w005",
    "n50w010",
    "n50w015",
    "n50w060",
    "n50w065",
    "n50w070",
    "n50w075",
    "n50w080",
    "n50w085",
    "n50w090",
    "n50w095",
    "n50w100",
    "n50w105",
    "n50w110",
    "n50w115",
    "n50w120",
    "n50w125",
    "n50w130",
    "n50w135",
    "n50w160",
    "n50w165",
    "n50w170",
    "n50w175",
    "n50w180",
    "n55e000",
    "n55e005",
    "n55e010",
    "n55e015",
    "n55e020",
    "n55e025",
    "n55e030",
    "n55e035",
    "n55e040",
    "n55e045",
    "n55e050",
    "n55e055",
    "n55e060",
    "n55e065",
    "n55e070",
    "n55e075",
    "n55e080",
    "n55e085",
    "n55e090",
    "n55e095",
    "n55e100",
    "n55e105",
    "n55e110",
    "n55e115",
    "n55e120",
    "n55e125",
    "n55e130",
    "n55e135",
    "n55e140",
    "n55e145",
    "n55e150",
    "n55e155",
    "n55e160",
    "n55e165",
    "n55e170",
    "n55w005",
    "n55w010",
    "n55w015",
    "n55w045",
    "n55w050",
    "n55w060",
    "n55w065",
    "n55w070",
    "n55w075",
    "n55w080",
    "n55w085",
    "n55w090",
    "n55w095",
    "n55w100",
    "n55w105",
    "n55w110",
    "n55w115",
    "n55w120",
    "n55w125",
    "n55w130",
    "n55w135",
    "n55w140",
    "n55w145",
    "n55w150",
    "n55w155",
    "n55w160",
    "n55w165",
    "n55w170",
    "n55w175",
    "n60e000",
    "n60e005",
    "n60e010",
    "n60e015",
    "n60e020",
    "n60e025",
    "n60e030",
    "n60e035",
    "n60e040",
    "n60e045",
    "n60e050",
    "n60e055",
    "n60e060",
    "n60e065",
    "n60e070",
    "n60e075",
    "n60e080",
    "n60e085",
    "n60e090",
    "n60e095",
    "n60e100",
    "n60e105",
    "n60e110",
    "n60e115",
    "n60e120",
    "n60e125",
    "n60e130",
    "n60e135",
    "n60e140",
    "n60e145",
    "n60e150",
    "n60e155",
    "n60e160",
    "n60e165",
    "n60e170",
    "n60e175",
    "n60w005",
    "n60w010",
    "n60w015",
    "n60w020",
    "n60w025",
    "n60w040",
    "n60w045",
    "n60w050",
    "n60w055",
    "n60w065",
    "n60w070",
    "n60w075",
    "n60w080",
    "n60w085",
    "n60w090",
    "n60w095",
    "n60w100",
    "n60w105",
    "n60w110",
    "n60w115",
    "n60w120",
    "n60w125",
    "n60w130",
    "n60w135",
    "n60w140",
    "n60w145",
    "n60w150",
    "n60w155",
    "n60w160",
    "n60w165",
    "n60w170",
    "n60w175",
    "n60w180",
    "n65e010",
    "n65e015",
    "n65e020",
    "n65e025",
    "n65e030",
    "n65e035",
    "n65e040",
    "n65e045",
    "n65e050",
    "n65e055",
    "n65e060",
    "n65e065",
    "n65e070",
    "n65e075",
    "n65e080",
    "n65e085",
    "n65e090",
    "n65e095",
    "n65e100",
    "n65e105",
    "n65e110",
    "n65e115",
    "n65e120",
    "n65e125",
    "n65e130",
    "n65e135",
    "n65e140",
    "n65e145",
    "n65e150",
    "n65e155",
    "n65e160",
    "n65e165",
    "n65e170",
    "n65e175",
    "n65w015",
    "n65w020",
    "n65w025",
    "n65w030",
    "n65w035",
    "n65w040",
    "n65w045",
    "n65w050",
    "n65w055",
    "n65w065",
    "n65w070",
    "n65w075",
    "n65w080",
    "n65w085",
    "n65w090",
    "n65w095",
    "n65w100",
    "n65w105",
    "n65w110",
    "n65w115",
    "n65w120",
    "n65w125",
    "n65w130",
    "n65w135",
    "n65w140",
    "n65w145",
    "n65w150",
    "n65w155",
    "n65w160",
    "n65w165",
    "n65w170",
    "n65w175",
    "n65w180",
    "n70e015",
    "n70e020",
    "n70e025",
    "n70e030",
    "n70e050",
    "n70e055",
    "n70e060",
    "n70e065",
    "n70e070",
    "n70e075",
    "n70e080",
    "n70e085",
    "n70e090",
    "n70e095",
    "n70e100",
    "n70e105",
    "n70e110",
    "n70e115",
    "n70e120",
    "n70e125",
    "n70e130",
    "n70e135",
    "n70e140",
    "n70e145",
    "n70e150",
    "n70e155",
    "n70e160",
    "n70e165",
    "n70e170",
    "n70e175",
    "n70w010",
    "n70w020",
    "n70w025",
    "n70w030",
    "n70w035",
    "n70w040",
    "n70w045",
    "n70w050",
    "n70w055",
    "n70w060",
    "n70w070",
    "n70w075",
    "n70w080",
    "n70w085",
    "n70w090",
    "n70w095",
    "n70w100",
    "n70w105",
    "n70w110",
    "n70w115",
    "n70w120",
    "n70w125",
    "n70w130",
    "n70w135",
    "n70w145",
    "n70w150",
    "n70w155",
    "n70w160",
    "n70w165",
    "n70w180",
    "n75e010",
    "n75e015",
    "n75e020",
    "n75e025",
    "n75e030",
    "n75e045",
    "n75e050",
    "n75e055",
    "n75e060",
    "n75e065",
    "n75e075",
    "n75e080",
    "n75e085",
    "n75e090",
    "n75e095",
    "n75e100",
    "n75e105",
    "n75e110",
    "n75e135",
    "n75e140",
    "n75e145",
    "n75e150",
    "n75e155",
    "n75w020",
    "n75w025",
    "n75w030",
    "n75w035",
    "n75w040",
    "n75w045",
    "n75w050",
    "n75w055",
    "n75w060",
    "n75w065",
    "n75w070",
    "n75w075",
    "n75w080",
    "n75w085",
    "n75w090",
    "n75w095",
    "n75w100",
    "n75w105",
    "n75w110",
    "n75w115",
    "n75w120",
    "n75w125",
    "n80e010",
    "n80e015",
    "n80e020",
    "n80e025",
    "n80e030",
    "n80e035",
    "n80e040",
    "n80e045",
    "n80e050",
    "n80e055",
    "n80e060",
    "n80e065",
    "n80e075",
    "n80e080",
    "n80e090",
    "n80e095",
    "n80w015",
    "n80w020",
    "n80w025",
    "n80w030",
    "n80w035",
    "n80w040",
    "n80w045",
    "n80w050",
    "n80w055",
    "n80w060",
    "n80w065",
    "n80w070",
    "n80w075",
    "n80w080",
    "n80w085",
    "n80w090",
    "n80w095",
    "n80w100",
    "n80w105",
    "s05e005",
    "s05e010",
    "s05e015",
    "s05e020",
    "s05e025",
    "s05e030",
    "s05e035",
    "s05e040",
    "s05e050",
    "s05e055",
    "s05e070",
    "s05e095",
    "s05e100",
    "s05e105",
    "s05e110",
    "s05e115",
    "s05e120",
    "s05e125",
    "s05e130",
    "s05e135",
    "s05e140",
    "s05e145",
    "s05e150",
    "s05e155",
    "s05e165",
    "s05e170",
    "s05e175",
    "s05w035",
    "s05w040",
    "s05w045",
    "s05w050",
    "s05w055",
    "s05w060",
    "s05w065",
    "s05w070",
    "s05w075",
    "s05w080",
    "s05w085",
    "s05w090",
    "s05w095",
    "s05w155",
    "s05w165",
    "s05w175",
    "s10e010",
    "s10e015",
    "s10e020",
    "s10e025",
    "s10e030",
    "s10e035",
    "s10e045",
    "s10e050",
    "s10e055",
    "s10e070",
    "s10e100",
    "s10e105",
    "s10e110",
    "s10e115",
    "s10e120",
    "s10e125",
    "s10e130",
    "s10e135",
    "s10e140",
    "s10e145",
    "s10e150",
    "s10e155",
    "s10e160",
    "s10e165",
    "s10e175",
    "s10w015",
    "s10w035",
    "s10w040",
    "s10w045",
    "s10w050",
    "s10w055",
    "s10w060",
    "s10w065",
    "s10w070",
    "s10w075",
    "s10w080",
    "s10w085",
    "s10w140",
    "s10w145",
    "s10w155",
    "s10w160",
    "s10w165",
    "s10w175",
    "s15e010",
    "s15e015",
    "s15e020",
    "s15e025",
    "s15e030",
    "s15e035",
    "s15e040",
    "s15e045",
    "s15e050",
    "s15e055",
    "s15e095",
    "s15e105",
    "s15e115",
    "s15e120",
    "s15e125",
    "s15e130",
    "s15e135",
    "s15e140",
    "s15e145",
    "s15e150",
    "s15e155",
    "s15e160",
    "s15e165",
    "s15e170",
    "s15e175",
    "s15w040",
    "s15w045",
    "s15w050",
    "s15w055",
    "s15w060",
    "s15w065",
    "s15w070",
    "s15w075",
    "s15w080",
    "s15w140",
    "s15w145",
    "s15w150",
    "s15w155",
    "s15w165",
    "s15w170",
    "s15w175",
    "s15w180",
    "s20e010",
    "s20e015",
    "s20e020",
    "s20e025",
    "s20e030",
    "s20e035",
    "s20e040",
    "s20e045",
    "s20e050",
    "s20e055",
    "s20e060",
    "s20e115",
    "s20e120",
    "s20e125",
    "s20e130",
    "s20e135",
    "s20e140",
    "s20e145",
    "s20e150",
    "s20e155",
    "s20e160",
    "s20e165",
    "s20e170",
    "s20e175",
    "s20w010",
    "s20w040",
    "s20w045",
    "s20w050",
    "s20w055",
    "s20w060",
    "s20w065",
    "s20w070",
    "s20w075",
    "s20w080",
    "s20w140",
    "s20w145",
    "s20w150",
    "s20w155",
    "s20w160",
    "s20w165",
    "s20w175",
    "s20w180",
    "s25e010",
    "s25e015",
    "s25e020",
    "s25e025",
    "s25e030",
    "s25e035",
    "s25e040",
    "s25e045",
    "s25e055",
    "s25e110",
    "s25e115",
    "s25e120",
    "s25e125",
    "s25e130",
    "s25e135",
    "s25e140",
    "s25e145",
    "s25e150",
    "s25e155",
    "s25e160",
    "s25e165",
    "s25e170",
    "s25w030",
    "s25w045",
    "s25w050",
    "s25w055",
    "s25w060",
    "s25w065",
    "s25w070",
    "s25w075",
    "s25w125",
    "s25w130",
    "s25w135",
    "s25w140",
    "s25w145",
    "s25w150",
    "s25w155",
    "s25w160",
    "s25w175",
    "s25w180",
    "s30e010",
    "s30e015",
    "s30e020",
    "s30e025",
    "s30e030",
    "s30e040",
    "s30e045",
    "s30e110",
    "s30e115",
    "s30e120",
    "s30e125",
    "s30e130",
    "s30e135",
    "s30e140",
    "s30e145",
    "s30e150",
    "s30e165",
    "s30w050",
    "s30w055",
    "s30w060",
    "s30w065",
    "s30w070",
    "s30w075",
    "s30w080",
    "s30w085",
    "s30w110",
    "s30w135",
    "s30w145",
    "s30w180",
    "s35e015",
    "s35e020",
    "s35e025",
    "s35e030",
    "s35e110",
    "s35e115",
    "s35e120",
    "s35e125",
    "s35e130",
    "s35e135",
    "s35e140",
    "s35e145",
    "s35e150",
    "s35e155",
    "s35e170",
    "s35w055",
    "s35w060",
    "s35w065",
    "s35w070",
    "s35w075",
    "s35w080",
    "s35w085",
    "s35w180",
    "s40e075",
    "s40e115",
    "s40e135",
    "s40e140",
    "s40e145",
    "s40e150",
    "s40e170",
    "s40e175",
    "s40w015",
    "s40w060",
    "s40w065",
    "s40w070",
    "s40w075",
    "s45e140",
    "s45e145",
    "s45e165",
    "s45e170",
    "s45e175",
    "s45w010",
    "s45w015",
    "s45w065",
    "s45w070",
    "s45w075",
    "s45w080",
    "s45w180",
    "s50e035",
    "s50e050",
    "s50e065",
    "s50e070",
    "s50e165",
    "s50e170",
    "s50e175",
    "s50w070",
    "s50w075",
    "s50w080",
    "s55e000",
    "s55e065",
    "s55e070",
    "s55e155",
    "s55e165",
    "s55w040",
    "s55w060",
    "s55w065",
    "s55w070",
    "s55w075",
    "s55w080",
    "s60w030",
    "s60w070",
    "s60w075",
}


class _ProgressReader:
    """Wrapper for file-like object to track read progress."""

    def __init__(self, fileobj: IO[bytes], pbar: tqdm) -> None:
        """Initialize the progress reader.

        Args:
            fileobj: File-like object to read from.
            pbar: tqdm progress bar to update.
        """
        self.fileobj: IO[bytes] = fileobj
        self.pbar: tqdm = pbar

    def read(self, size: int = -1) -> bytes:
        """Read data and update progress bar.

        Args:
            size: Read size in bytes. If -1, read all available. Defaults to -1.

        Returns:
            Data read from the file-like object.
        """
        data = self.fileobj.read(size)
        if data:
            self.pbar.update(len(data))
        return data

    def close(self) -> None:
        self.fileobj.close()


class MeritHydro(Adapter):
    """Dataset adapter for MERIT Hydro variables."""

    _xmin: float
    _xmax: float
    _ymin: float
    _ymax: float
    _source_nodata: int | float | bool
    _target_nodata: int | float | bool

    def __init__(self, variable: str, *args: Any, **kwargs: Any) -> None:
        """Initialize the adapter for a specific MERIT Hydro variable.

        Args:
            variable: MERIT Hydro variable to download ("elv" or "dir").

            *args: Additional positional arguments passed to the base Adapter class.
            **kwargs: Additional keyword arguments passed to the base Adapter class.
        """
        self.variable = variable
        super().__init__(*args, **kwargs)

    @property
    def is_ready(self) -> bool:
        """Check if the data is already downloaded and processed.

        For MERIT Hydro, we always return False because readiness depends
        on the required tiles for a specific bounding box, which are checked
        during fetch().

        Returns:
            Always False to ensure fetch() is called.
        """
        return False

    def _get_latitude_hemisphere_and_degrees(self, lat_deg: int) -> tuple[str, int]:
        """Return hemisphere letter and absolute degrees for latitude.

        Args:
            lat_deg: Integer degrees latitude (lower-left corner of tile) (degrees).

        Returns:
            Tuple of hemisphere letter ("n" or "s") and two-digit absolute degrees.

        Raises:
            ValueError: If latitude is outside valid range [-90, 90].
        """
        if lat_deg < -90 or lat_deg > 90:
            raise ValueError("Latitude must be within [-90, 90] degrees.")
        if lat_deg >= 0:
            return "n", lat_deg
        return "s", -lat_deg

    def _get_longitude_hemisphere_and_degrees(self, lon_deg: int) -> tuple[str, int]:
        """Return hemisphere letter and absolute degrees for longitude.

        Args:
            lon_deg: Integer degrees longitude (lower-left corner of tile) (degrees).

        Returns:
            Tuple of hemisphere letter ("e" or "w") and three-digit absolute degrees.

        Raises:
            ValueError: If longitude is outside valid range [-180, 180].
        """
        if lon_deg < -180 or lon_deg > 180:
            raise ValueError("Longitude must be within [-180, 180] degrees.")
        if lon_deg >= 0:
            return "e", lon_deg
        return "w", -lon_deg

    def _compose_tile_filename(self, lat_ll: int, lon_ll: int) -> str:
        """Compose a 5x5-degree tile filename for a MERIT variable.

        Args:
            lat_ll: Lower-left latitude of tile (integer multiple of 5) (degrees).
            lon_ll: Lower-left longitude of tile (integer multiple of 5) (degrees).

        Returns:
            Tile filename like "n30w120_elv.tif".

        Raises:
            ValueError: If lat_ll or lon_ll are not multiples of 5.
        """
        if lat_ll % 5 != 0 or lon_ll % 5 != 0:
            raise ValueError(
                "lat_ll and lon_ll must be 5-degree aligned (multiples of 5)."
            )
        ns, alat = self._get_latitude_hemisphere_and_degrees(lat_ll)
        ew, alon = self._get_longitude_hemisphere_and_degrees(lon_ll)
        return f"{ns}{alat:02d}{ew}{alon:03d}_{self.variable}.tif"

    def _package_name(self, lat_ll: int, lon_ll: int) -> str:
        """Compose a 30x30-degree package tar filename for a MERIT variable.

        The package is defined by the lower-left corner of the 30-degree grid cell
        that contains the tile lower-left corner.

        Args:
            lat_ll: Lower-left latitude of tile (integer multiple of 5) (degrees).
            lon_ll: Lower-left longitude of tile (integer multiple of 5) (degrees).

        Returns:
            Package tar filename like "elv_n30w120.tar".
        """
        # Floor to 30-degree grid
        lat30 = (lat_ll // 30) * 30 if lat_ll >= 0 else -(((-lat_ll + 29) // 30) * 30)
        lon30 = (lon_ll // 30) * 30 if lon_ll >= 0 else -(((-lon_ll + 29) // 30) * 30)
        ns, alat = self._get_latitude_hemisphere_and_degrees(lat30)
        ew, alon = self._get_longitude_hemisphere_and_degrees(lon30)
        return f"{self.variable}_{ns}{alat:02d}{ew}{alon:03d}.tar"

    def _tiles_for_bbox(
        self, xmin: float, xmax: float, ymin: float, ymax: float
    ) -> list[tuple[int, int]]:
        """Compute all 5x5-degree lower-left tile coordinates intersecting a bbox.

        Args:
            xmin: Minimum longitude (degrees).
            xmax: Maximum longitude (degrees). May be less than xmin if crossing dateline is needed (not supported here).
            ymin: Minimum latitude (degrees).
            ymax: Maximum latitude (degrees).

        Returns:
            Sorted list of (lat_ll, lon_ll) integer tuples on 5-degree grid.

        Raises:
            ValueError: If bbox is invalid or crosses the antimeridian.
        """
        if xmax <= xmin:
            raise ValueError("Expected xmax > xmin and bbox not crossing antimeridian.")
        if ymax <= ymin:
            raise ValueError("Expected ymax > ymin.")

        # Clamp to plausible world bounds to avoid generating excessive tiles
        xmin_c = max(-180.0, xmin)
        xmax_c = min(180.0, xmax)
        ymin_c = max(-90.0, ymin)
        ymax_c = min(90.0, ymax)

        # Align to 5-degree grid (lower-left corners)
        def floor5(v: float) -> int:
            vi = int(v // 5 * 5)
            # Adjust for negatives that are exact multiples to ensure lower-left
            if v < 0 and v % 5 == 0:
                return vi
            return vi

        lat_start = floor5(ymin_c)
        lon_start = floor5(xmin_c)

        tiles: list[tuple[int, int]] = []
        lat = lat_start
        while lat < ymax_c:
            lon = lon_start
            while lon < xmax_c:
                # Tile bounds
                t_xmin, t_xmax = lon, lon + 5
                t_ymin, t_ymax = lat, lat + 5
                if not (
                    t_xmin >= xmax_c
                    or t_xmax <= xmin_c
                    or t_ymin >= ymax_c
                    or t_ymax <= ymin_c
                ):
                    tiles.append((lat, lon))
                lon += 5
            lat += 5
        # Unique and sorted for reproducibility
        tiles = sorted(set(tiles))
        return tiles

    def _group_tiles_by_package(
        self, tiles: Iterable[tuple[int, int]]
    ) -> dict[str, list[tuple[int, int]]]:
        """Group tile ll coords by the 30x30 package tar they belong to.

        Args:
            tiles: Iterable of (lat_ll, lon_ll) pairs (degrees).

        Returns:
            Mapping of package tar filename -> list of tile coords in that package.
        """
        groups: dict[str, list[tuple[int, int]]] = {}
        for lat_ll, lon_ll in tiles:
            pkg = self._package_name(lat_ll, lon_ll)
            groups.setdefault(pkg, []).append((lat_ll, lon_ll))
        return groups

    def _package_url(self, package_name: str, base_url: str) -> str:
        """Construct the full URL to a MERIT tar package for the given variable.

        Args:
            package_name: Tar filename produced by _package_name.
            base_url: Base URL for MERIT Hydro downloads.

        Returns:
            Full URL string to the tar file.
        """
        return f"{base_url}/{package_name}"

    def _merge_merit_tiles(self, tile_paths: list[Path]) -> xr.DataArray:
        """Load MERIT Hydro tiles into a single xarray DataArray.

        This function opens the provided GeoTIFF tile files using rioxarray,
        merges them into a single DataArray, and returns it in memory.

        Args:
            tile_paths: List of Paths to GeoTIFF files from download_merit.

        Returns:
            xarray DataArray with merged tiles, preserving CRS and coordinates.

        Raises:
            ValueError: If data types among tiles are inconsistent.
        """
        das: list[xr.DataArray] = []
        for path in tile_paths:
            src = rxr.open_rasterio(path)
            assert isinstance(src, xr.DataArray)
            das.append(src.sel(band=1))

        first_dtype = das[0].dtype
        for da in das:
            if da.dtype != first_dtype:
                raise ValueError("Inconsistent data types among tiles.")

        # there is an issue with merging uint8 tiles. Therefore, we convert to int32
        # before merging and then convert back.
        if das[0].dtype == np.uint8:
            das = [da.astype(np.int32) for da in das]
            is_uint8 = True
        else:
            is_uint8 = False

        da: xr.DataArray = merge.merge_arrays(das, nodata=self._source_nodata)
        if is_uint8:
            da = da.astype(np.uint8)
        return da

    def _missing_marker_path(self, tile_name: str) -> Path:
        """Construct path for a missing tile marker file.

        Args:
            tile_name: Name of the tile.

        Returns:
            Path to the marker file.
        """
        root = self.root
        assert root is not None
        return root / f"{tile_name}.missing.txt"

    def fetch(
        self,
        *,
        xmin: float,
        xmax: float,
        ymin: float,
        ymax: float,
        url: str,
        source_nodata: int | float | bool,
        target_nodata: int | float | bool,
        session: requests.Session | None = None,
        request_timeout_s: float = 60.0,
        attempts: int = 3,
    ) -> MeritHydro:
        """Ensure MERIT Hydro tiles intersecting a bbox are available locally.

        The function first checks for pre-staged 5x5-degree tiles in
        ``{cache_root}/{variable}/`` or ``{cache_root}/``. If tiles are
        missing, it downloads only the required GeoTIFFs for a single MERIT variable
        by streaming the remote 30x30-degree tar packages without saving the tars.
        If a package does not exist (HTTP 404), it is silently skipped. If a needed
        tile is not present inside an existing package (commonly ocean), it is also
        silently skipped. If a tile should be on disk but is missing, it tries to download it.
        Any other error is retried up to ``attempts`` times and
        then raised.

        Authentication:
            Basic auth is required and read from environment variables:
            MERIT_USERNAME and MERIT_PASSWORD.

        Args:
            xmin: Minimum longitude of area of interest (degrees).
            xmax: Maximum longitude of area of interest (degrees).
            ymin: Minimum latitude of area of interest (degrees).
            ymax: Maximum latitude of area of interest (degrees).
            url: Base URL of the MERIT Hydro server.
            source_nodata: Nodata value in the source GeoTIFF files.
            target_nodata: Nodata value to use in the output DataArray.
            session: Optional requests.Session (used in tests or advanced usage).
            request_timeout_s: Timeout per HTTP request (seconds).
            attempts: Number of attempts for transient failures (errors are raised after this many tries).

        Returns:
            The MeritHydro instance.

        Raises:
            ValueError: If inputs are invalid or auth variables are missing.
            RuntimeError: If the HTTP client dependency is not available.
            requests.RequestException: If repeated HTTP errors occur.
            tarfile.ReadError: If tar parsing repeatedly fails.
        """
        self._xmin = xmin
        self._xmax = xmax
        self._ymin = ymin
        self._ymax = ymax
        self._source_nodata = source_nodata
        self._target_nodata = target_nodata

        username = os.getenv("MERIT_USERNAME")
        password = os.getenv("MERIT_PASSWORD")
        if not username or not password:
            raise ValueError(
                "Authentication required: set MERIT_USERNAME and MERIT_PASSWORD in environment."
            )

        tiles: list[tuple[int, int]] = self._tiles_for_bbox(xmin, xmax, ymin, ymax)

        local_tile_dir: Path = self.root / self.variable
        missing_names: set[str] = set()
        for lat_ll, lon_ll in tiles:
            tile_name: str = self._compose_tile_filename(lat_ll, lon_ll)
            if (
                not (self.root / tile_name).exists()
                and not (local_tile_dir / tile_name).exists()
                and tile_name[:7] in available_tiles
            ):
                if not self._missing_marker_path(tile_name).exists():
                    missing_names.add(tile_name)

        if not missing_names:
            return self

        # Prepare HTTP session
        if session is None:
            if requests is None:
                raise RuntimeError(
                    "requests is required for downloading but is not available."
                )
            session = requests.Session()
        auth = HTTPBasicAuth(username, password)

        missing_coords = []
        for lat_ll, lon_ll in tiles:
            if self._compose_tile_filename(lat_ll, lon_ll) in missing_names:
                missing_coords.append((lat_ll, lon_ll))

        groups = self._group_tiles_by_package(missing_coords)

        for package_name, coords in groups.items():
            package_url = self._package_url(package_name, base_url=url)
            needed_names = {
                self._compose_tile_filename(lat, lon) for lat, lon in coords
            }

            # HEAD to detect ocean/no-data packages (404) → skip silently
            try:
                head_resp = session.head(
                    package_url,
                    auth=auth,
                    allow_redirects=True,
                    timeout=request_timeout_s,
                )
            except Exception:
                if attempts <= 1:
                    raise
                retried = 1
                while retried < attempts:
                    time.sleep(min(2.0 * retried, 5.0))
                    try:
                        head_resp = session.head(
                            package_url,
                            auth=auth,
                            allow_redirects=True,
                            timeout=request_timeout_s,
                        )
                        break
                    except Exception:
                        retried += 1
                else:
                    raise

            if head_resp.status_code == 404:
                for tname in needed_names:
                    mm = self._missing_marker_path(tname)
                    mm.write_text(
                        "MERIT Hydro: tile not provided (ocean/no-data).\n",
                        encoding="utf-8",
                    )
                continue
            if head_resp.status_code in (401, 403):
                raise requests.RequestException(
                    f"Unauthorized: HTTP {head_resp.status_code} for {package_url}"
                )
            if head_resp.status_code >= 400:
                raise requests.RequestException(
                    f"HEAD error: HTTP {head_resp.status_code} for {package_url}"
                )

            content_length = int(head_resp.headers.get("Content-Length", 0))
            found_names: set[str] = set()
            for attempt in range(1, attempts + 1):
                try:
                    resp = session.get(
                        package_url, auth=auth, stream=True, timeout=request_timeout_s
                    )
                except Exception:
                    if attempt >= attempts:
                        raise
                    time.sleep(min(2.0 * attempt, 5.0))
                    continue

                if resp.status_code >= 400:
                    resp.close()
                    if attempt >= attempts:
                        raise requests.RequestException(
                            f"GET error: HTTP {resp.status_code} for {package_url}"
                        )
                    time.sleep(min(2.0 * attempt, 5.0))
                    continue

                try:
                    resp.raw.decode_content = True
                    print(f"Downloading and extracting from: {package_url}")
                    with tqdm(
                        total=content_length if content_length > 0 else None,
                        desc=f"Downloading {package_name}",
                        unit="B",
                        unit_scale=True,
                    ) as pbar:
                        progress_reader = _ProgressReader(resp.raw, pbar)
                        with tarfile.open(fileobj=progress_reader, mode="r|*") as tf:  # type: ignore
                            for member in tf:
                                if not member.isreg():
                                    continue
                                mname = Path(member.name).name
                                if mname in needed_names:
                                    ex = tf.extractfile(member)
                                    if ex is None:
                                        continue
                                    out_file = self.root / mname
                                    with out_file.open("wb") as fout:
                                        while True:
                                            chunk = ex.read(1024 * 1024)
                                            if not chunk:
                                                break
                                            fout.write(chunk)
                                    found_names.add(mname)
                                    if found_names == needed_names:
                                        break
                except tarfile.ReadError:
                    if attempt >= attempts:
                        raise
                    time.sleep(min(2.0 * attempt, 5.0))
                    continue
                finally:
                    resp.close()
                break

            for tname in needed_names - found_names:
                mm = self._missing_marker_path(tname)
                mm.write_text(
                    "MERIT Hydro: tile not present in package (likely ocean).\n",
                    encoding="utf-8",
                )

        return self

    def read(self, **kwargs: Any) -> xr.DataArray:
        """Read and merge the MERIT Hydro tiles into a single DataArray.

        This method should be called after fetch(). It finds all available tiles
        for the stored bounding box, merges them, crops to the exact bbox,
        and applies nodata conversion.

        Args:
            **kwargs: Additional keyword arguments (ignored).

        Returns:
            Merged and cropped xarray DataArray.

        Raises:
            ValueError: If fetch() has not been called.
        """
        if (
            not hasattr(self, "_xmin")
            or not hasattr(self, "_xmax")
            or not hasattr(self, "_ymin")
            or not hasattr(self, "_ymax")
            or not hasattr(self, "_source_nodata")
            or not hasattr(self, "_target_nodata")
        ):
            raise ValueError(
                "fetch() must be called before read() to set the bounding box."
            )

        tiles: list[tuple[int, int]] = self._tiles_for_bbox(
            self._xmin, self._xmax, self._ymin, self._ymax
        )
        local_tile_dir: Path = self.root / self.variable
        results: list[Path] = []
        for lat_ll, lon_ll in tiles:
            tname = self._compose_tile_filename(lat_ll, lon_ll)
            tif_path = self.root / tname
            if tif_path.exists():
                results.append(tif_path)
                continue
            local_path = local_tile_dir / tname
            if local_path.exists():
                results.append(local_path)
                continue

        if not results:
            raise ValueError(
                f"No MERIT Hydro tiles found for bbox ({self._xmin}, {self._ymin}, {self._xmax}, {self._ymax})."
            )

        da: xr.DataArray = self._merge_merit_tiles(results)

        if "_FillValue" in da.attrs:
            assert da.attrs["_FillValue"] == self._source_nodata, (
                f"Expected source _FillValue {self._source_nodata}, got {da.attrs['_FillValue']}"
            )
        else:
            da.attrs["_FillValue"] = self._source_nodata

        da = da.sel(x=slice(self._xmin, self._xmax), y=slice(self._ymax, self._ymin))
        da = convert_nodata(da, self._target_nodata)
        return da


class MeritHydroDir(MeritHydro):
    """Dataset adapter for MERIT Hydro flow direction.

    Args:
        MeritHydro: Base class for MERIT Hydro datasets.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the adapter for flow direction data.

        Args:
            *args: Positional arguments passed to the base class.
            **kwargs: Keyword arguments passed to the base class.

        """
        super().__init__(variable="dir", **kwargs)

    def fetch(self, **kwargs: Any) -> MeritHydro:
        """Process and download flow direction data with specific fill value.

        Args:
            **kwargs: Keyword arguments passed to the base class fetcher.

        Returns:
            The MeritHydro instance.

        """
        return super().fetch(
            source_nodata=247,
            target_nodata=247,
            **kwargs,
        )


class MeritHydroElv(MeritHydro):
    """Dataset adapter for MERIT Hydro elevation.

    Args:
        MeritHydro: Base class for MERIT Hydro datasets.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the adapter for elevation data.

        Args:
            *args: Positional arguments passed to the base class.
            **kwargs: Keyword arguments passed to the base class.

        """
        super().__init__(variable="elv", **kwargs)

    def fetch(self, **kwargs: Any) -> MeritHydro:
        """Process and download elevation data with specific fill value.

        Args:
            **kwargs: Keyword arguments passed to the base class fetcher.

        Returns:
            The MeritHydro instance.

        """
        return super().fetch(
            source_nodata=-9999.0,
            target_nodata=np.nan,
            **kwargs,
        )



"""OpenStreetMap data catalog adapter and workflow."""

from __future__ import annotations

import math
import tempfile
from pathlib import Path
from typing import Any

import geopandas as gpd
import pandas as pd
import requests
from shapely.geometry.polygon import Polygon
from tqdm import tqdm

from geb.workflows.io import fetch_and_save

from .base import Adapter


class OpenStreetMap(Adapter):
    """Open Street Map data catalog adapter."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the Open StreetMap adapter."""
        super().__init__(*args, **kwargs)

    def fetch(self, url: str) -> OpenStreetMap:
        """Sets the URL for the OSM data source.

        Args:
            url: The URL template for the OSM dataset.

        Returns:
            The OpenStreetMap adapter instance.
        """
        self.url = url
        return self

    def filter_regions(self, ID: str, parents: list[str]) -> bool:
        """Filter out regions that are children of other regions in the list.

        Args:
            ID: The ID of the region to check.
            parents: The list of parent region IDs.

        Returns:
            bool: True if the region is not a child of any other region in the list, False otherwise.
        """
        return ID not in parents

    def read(
        self, geom: Polygon, feature_types: list[str]
    ) -> dict[str, gpd.GeoDataFrame]:
        """Download and read OSM data for the specified geometry and feature types.

        First, the method identifies all regions in the OSM index that intersect with the provided geometry.
        It then filters out any regions that are children of other intersecting regions to avoid redundant data
        downloads. For each of the remaining regions, it downloads the corresponding OSM data file and
        extracts the specified feature types (e.g., buildings, roads, rails) using geopandas. The extracted features
        from all regions are concatenated into a single GeoDataFrame for each feature type.

        Args:
            geom: The geometry to read OSM data for.
            feature_types: The list of feature types to read (e.g., 'buildings', 'roads', 'rails').

        Returns:
            A dictionary with feature types as keys and GeoDataFrames as values.

        Raises:
            ValueError: If an unknown feature type is specified.
        """
        minx, miny, maxx, maxy = geom.bounds

        urls: list[str] = []
        for x in range(math.floor(minx), math.ceil(maxx)):
            EW_code = f"E{-x:03d}" if x <= 0 else f"W{x:03d}"
            for y in range(math.floor(miny), math.ceil(maxy)):
                NS_code = f"N{y:02d}" if y >= 0 else f"S{-y:02d}"
                # Create a polygon for the tile's bounding box (1x1 degree)
                tile_poly = Polygon([(x, y), (x + 1, y), (x + 1, y + 1), (x, y + 1)])
                # Only proceed if the tile intersects with the geometry
                if tile_poly.intersects(geom):
                    url = f"{self.url}/grid/{NS_code}{EW_code}-latest.osm.pbf"
                    # Some tiles do not exist because they are in the ocean. Therefore we check if they exist
                    # before adding the url
                    response = requests.head(url, allow_redirects=True)
                    if response.status_code != 404:
                        urls.append(url)

        all_features_list: dict[str, list[gpd.GeoDataFrame]] = {}

        print(
            f"Downloading and processing OSM data for {len(urls)} tiles (note: some tiles are larger than others)..."
        )
        for url in tqdm(urls):
            with tempfile.NamedTemporaryFile(suffix=".pbf") as tmp:
                filepath: Path = Path(tmp.name)

                fetch_and_save(
                    url, filepath, overwrite=True, show_progress=False, verbose=False
                )

                for feature_type in feature_types:
                    if feature_type not in all_features_list:
                        all_features_list[feature_type] = []

                    if feature_type == "buildings":
                        features: gpd.GeoDataFrame = gpd.read_file(
                            filepath,
                            mask=geom,
                            layer="multipolygons",
                            use_arrow=True,
                        )
                        features = features[features["building"].notna()]
                    elif feature_type == "rails":
                        features: gpd.GeoDataFrame = gpd.read_file(
                            filepath,
                            mask=geom,
                            layer="lines",
                            use_arrow=True,
                        )
                        features = features[
                            features["railway"].isin(
                                ["rail", "tram", "subway", "light_rail", "narrow_gauge"]
                            )
                        ]
                    elif feature_type == "roads":
                        features: gpd.GeoDataFrame = gpd.read_file(
                            filepath,
                            mask=geom,
                            layer="lines",
                            use_arrow=True,
                        )
                        features: gpd.GeoDataFrame = features[
                            features["highway"].isin(
                                [
                                    "motorway",
                                    "trunk",
                                    "primary",
                                    "secondary",
                                    "tertiary",
                                    "unclassified",
                                    "residential",
                                    "motorway_link",
                                    "trunk_link",
                                    "primary_link",
                                    "secondary_link",
                                    "tertiary_link",
                                ]
                            )
                        ]
                    else:
                        raise ValueError(f"Unknown feature type {feature_type}")

                    all_features_list[feature_type].append(features)

        all_features: dict[str, gpd.GeoDataFrame] = {}
        for feature_type in feature_types:
            all_features[feature_type] = pd.concat(
                all_features_list[feature_type], ignore_index=True
            )  # ty:ignore[invalid-assignment]

        return all_features
